# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import aqt
import aqt.forms
import aqt.operations
from anki.collection import Collection
from anki.consts import *
from anki.decks import DeckId
from anki.scheduler import CustomStudyRequest
from anki.scheduler.base import CustomStudyDefaults
from aqt.operations import QueryOp
from aqt.operations.scheduling import custom_study
from aqt.qt import *
from aqt.taglimit import TagLimit
from aqt.utils import add_close_shortcut, disable_help_button, tr

RADIO_NEW = 1
RADIO_REV = 2
RADIO_FORGOT = 3
RADIO_AHEAD = 4
RADIO_PREVIEW = 5
RADIO_CRAM = 6

TYPE_NEW = 0
TYPE_DUE = 1
TYPE_REVIEW = 2
TYPE_ALL = 3


class CustomStudy(QDialog):
    @staticmethod
    def fetch_data_and_show(mw: aqt.AnkiQt) -> None:
        def fetch_data(
            col: Collection,
        ) -> tuple[DeckId, CustomStudyDefaults]:
            deck_id = mw.col.decks.get_current_id()
            defaults = col.sched.custom_study_defaults(deck_id)
            return (deck_id, defaults)

        def show_dialog(data: tuple[DeckId, CustomStudyDefaults]) -> None:
            deck_id, defaults = data
            CustomStudy(mw=mw, deck_id=deck_id, defaults=defaults)

        QueryOp(
            parent=mw, op=fetch_data, success=show_dialog
        ).with_progress().run_in_background()

    def __init__(
        self,
        mw: aqt.AnkiQt,
        deck_id: DeckId,
        defaults: CustomStudyDefaults,
    ) -> None:
        "Don't call this directly; use CustomStudy.fetch_data_and_show()."
        QDialog.__init__(self, mw)
        self.mw = mw
        self.deck_id = deck_id
        self.defaults = defaults
        self.form = aqt.forms.customstudy.Ui_Dialog()
        self.form.setupUi(self)
        disable_help_button(self)
        self.setupSignals()
        add_close_shortcut(self)
        self.form.radioNew.click()
        self.open()

    def setupSignals(self) -> None:
        f = self.form
        qconnect(f.radioNew.clicked, lambda: self.onRadioChange(RADIO_NEW))
        qconnect(f.radioRev.clicked, lambda: self.onRadioChange(RADIO_REV))
        qconnect(f.radioForgot.clicked, lambda: self.onRadioChange(RADIO_FORGOT))
        qconnect(f.radioAhead.clicked, lambda: self.onRadioChange(RADIO_AHEAD))
        qconnect(f.radioPreview.clicked, lambda: self.onRadioChange(RADIO_PREVIEW))
        qconnect(f.radioCram.clicked, lambda: self.onRadioChange(RADIO_CRAM))

    def count_with_children(self, parent: int, children: int) -> str:
        if children:
            return f"{parent} {tr.custom_study_available_child_count(children)}"
        else:
            return str(parent)

    def onRadioChange(self, idx: int) -> None:
        form = self.form
        min_spinner_value = 1
        max_spinner_value = DYN_MAX_SIZE
        current_spinner_value = 1
        text_after_spinner = tr.custom_study_cards()
        title_text = ""
        show_cram_type = False
        ok = tr.custom_study_ok()

        if idx == RADIO_NEW:
            title_text = tr.custom_study_available_new_cards_2(
                count_string=self.count_with_children(
                    self.defaults.available_new,
                    self.defaults.available_new_in_children,
                ),
            )
            text_before_spinner = tr.custom_study_increase_todays_new_card_limit_by()
            current_spinner_value = self.defaults.extend_new
            min_spinner_value = -DYN_MAX_SIZE
        elif idx == RADIO_REV:
            title_text = tr.custom_study_available_review_cards_2(
                count_string=self.count_with_children(
                    self.defaults.available_review,
                    self.defaults.available_review_in_children,
                ),
            )
            text_before_spinner = tr.custom_study_increase_todays_review_limit_by()
            current_spinner_value = self.defaults.extend_review
            min_spinner_value = -DYN_MAX_SIZE
        elif idx == RADIO_FORGOT:
            text_before_spinner = tr.custom_study_review_cards_forgotten_in_last()
            text_after_spinner = tr.scheduling_days()
            max_spinner_value = 30
        elif idx == RADIO_AHEAD:
            text_before_spinner = tr.custom_study_review_ahead_by()
            text_after_spinner = tr.scheduling_days()
        elif idx == RADIO_PREVIEW:
            text_before_spinner = tr.custom_study_preview_new_cards_added_in_the()
            text_after_spinner = tr.scheduling_days()
            current_spinner_value = 1
        elif idx == RADIO_CRAM:
            text_before_spinner = tr.custom_study_select()
            text_after_spinner = tr.custom_study_cards_from_the_deck()
            ok = tr.custom_study_choose_tags()
            current_spinner_value = 100
            show_cram_type = True
        else:
            assert 0

        form.spin.setVisible(True)
        form.cardType.setVisible(show_cram_type)
        form.title.setText(title_text)
        form.title.setVisible(bool(title_text))
        form.spin.setMinimum(min_spinner_value)
        form.spin.setMaximum(max_spinner_value)
        if max_spinner_value > 0:
            form.spin.setEnabled(True)
        else:
            form.spin.setEnabled(False)
        form.spin.setValue(current_spinner_value)
        form.preSpin.setText(text_before_spinner)
        form.postSpin.setText(text_after_spinner)

        ok_button = form.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        assert ok_button is not None
        ok_button.setText(ok)

        self.radioIdx = idx

    def accept(self) -> None:
        request = CustomStudyRequest(deck_id=self.deck_id)
        if self.radioIdx == RADIO_NEW:
            request.new_limit_delta = self.form.spin.value()
        elif self.radioIdx == RADIO_REV:
            request.review_limit_delta = self.form.spin.value()
        elif self.radioIdx == RADIO_FORGOT:
            request.forgot_days = self.form.spin.value()
        elif self.radioIdx == RADIO_AHEAD:
            request.review_ahead_days = self.form.spin.value()
        elif self.radioIdx == RADIO_PREVIEW:
            request.preview_days = self.form.spin.value()
        else:
            request.cram.card_limit = self.form.spin.value()

            cram_type = self.form.cardType.currentRow()
            if cram_type == TYPE_NEW:
                request.cram.kind = CustomStudyRequest.Cram.CRAM_KIND_NEW
            elif cram_type == TYPE_DUE:
                request.cram.kind = CustomStudyRequest.Cram.CRAM_KIND_DUE
            elif cram_type == TYPE_REVIEW:
                request.cram.kind = CustomStudyRequest.Cram.CRAM_KIND_REVIEW
            else:
                request.cram.kind = CustomStudyRequest.Cram.CRAM_KIND_ALL

            def on_done(include: list[str], exclude: list[str]) -> None:
                request.cram.tags_to_include.extend(include)
                request.cram.tags_to_exclude.extend(exclude)
                self._create_and_close(request)

            # continues in background
            TagLimit(self, self.defaults.tags, on_done)
            return

        # other cases are synchronous
        self._create_and_close(request)

    def _create_and_close(self, request: CustomStudyRequest) -> None:
        # keep open on failure, as the cause was most likely an empty search
        # result, which the user can remedy
        custom_study(parent=self, request=request).success(
            lambda _: QDialog.accept(self)
        ).run_in_background()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import os
import re
import time
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass

import aqt.forms
import aqt.main
from anki.collection import (
    DeckIdLimit,
    ExportAnkiPackageOptions,
    ExportLimit,
    NoteIdsLimit,
    Progress,
)
from anki.decks import DeckId, DeckNameId
from anki.notes import NoteId
from aqt import gui_hooks
from aqt.errors import show_exception
from aqt.operations import QueryOp
from aqt.progress import ProgressUpdate
from aqt.qt import *
from aqt.utils import (
    add_close_shortcut,
    checkInvalidFilename,
    disable_help_button,
    getSaveFile,
    showWarning,
    tooltip,
    tr,
)


class ExportDialog(QDialog):
    def __init__(
        self,
        mw: aqt.main.AnkiQt,
        did: DeckId | None = None,
        nids: Sequence[NoteId] | None = None,
        parent: QWidget | None = None,
    ):
        QDialog.__init__(self, parent or mw, Qt.WindowType.Window)
        self.mw = mw
        self.col = mw.col.weakref()
        self.frm = aqt.forms.exporting.Ui_ExportDialog()
        self.frm.setupUi(self)
        self.exporter: Exporter
        self.nids = nids
        disable_help_button(self)
        self.setup(did)
        add_close_shortcut(self)
        self.open()

    def setup(self, did: DeckId | None) -> None:
        self.exporter_classes: list[type[Exporter]] = [
            ApkgExporter,
            ColpkgExporter,
            NoteCsvExporter,
            CardCsvExporter,
        ]
        gui_hooks.exporters_list_did_initialize(self.exporter_classes)
        self.frm.format.insertItems(
            0, [f"{e.name()} (.{e.extension})" for e in self.exporter_classes]
        )
        qconnect(self.frm.format.activated, self.exporter_changed)
        if self.nids is None and not did:
            # file>export defaults to colpkg
            default_exporter_idx = 1
        else:
            default_exporter_idx = 0
        self.frm.format.setCurrentIndex(default_exporter_idx)
        self.exporter_changed(default_exporter_idx)
        # deck list
        if self.nids is None:
            self.all_decks = self.col.decks.all_names_and_ids()
            decks = [tr.exporting_all_decks()]
            decks.extend(d.name for d in self.all_decks)
        else:
            decks = [tr.exporting_selected_notes()]
        self.frm.deck.addItems(decks)
        # save button
        b = QPushButton(tr.exporting_export())
        self.frm.buttonBox.addButton(b, QDialogButtonBox.ButtonRole.AcceptRole)
        self.frm.includeHTML.setChecked(True)
        # set default option if accessed through deck button
        if did:
            deck = self.mw.col.decks.get(did)
            assert deck is not None
            name = deck["name"]
            index = self.frm.deck.findText(name)
            self.frm.deck.setCurrentIndex(index)
            self.frm.includeSched.setChecked(False)

    def exporter_changed(self, idx: int) -> None:
        self.exporter = self.exporter_classes[idx]()
        self.frm.includeSched.setVisible(self.exporter.show_include_scheduling)
        self.frm.include_deck_configs.setVisible(
            self.exporter.show_include_deck_configs
        )
        self.frm.includeMedia.setVisible(self.exporter.show_include_media)
        self.frm.includeTags.setVisible(self.exporter.show_include_tags)
        self.frm.includeHTML.setVisible(self.exporter.show_include_html)
        self.frm.includeDeck.setVisible(self.exporter.show_include_deck)
        self.frm.includeNotetype.setVisible(self.exporter.show_include_notetype)
        self.frm.includeGuid.setVisible(self.exporter.show_include_guid)
        self.frm.legacy_support.setVisible(self.exporter.show_legacy_support)
        self.frm.deck.setVisible(self.exporter.show_deck_list)

    def accept(self) -> None:
        if not (out_path := self.get_out_path()):
            return
        self.exporter.export(self.mw, self.options(out_path))
        QDialog.reject(self)

    def get_out_path(self) -> str | None:
        filename = self.filename()
        while True:
            path = getSaveFile(
                parent=self,
                title=tr.actions_export(),
                dir_description="export",
                key=self.exporter.name(),
                ext="." + self.exporter.extension,
                fname=filename,
            )
            if not path:
                return None
            if checkInvalidFilename(os.path.basename(path), dirsep=False):
                continue
            path = os.path.normpath(path)
            if os.path.commonprefix([self.mw.pm.base, path]) == self.mw.pm.base:
                showWarning("Please choose a different export location.")
                continue
            break
        return path

    def options(self, out_path: str) -> ExportOptions:
        limit: ExportLimit | None = None
        if self.nids:
            limit = NoteIdsLimit(self.nids)
        elif current_deck_id := self.current_deck_id():
            limit = DeckIdLimit(current_deck_id)

        return ExportOptions(
            out_path=out_path,
            include_scheduling=self.frm.includeSched.isChecked(),
            include_deck_configs=self.frm.include_deck_configs.isChecked(),
            include_media=self.frm.includeMedia.isChecked(),
            include_tags=self.frm.includeTags.isChecked(),
            include_html=self.frm.includeHTML.isChecked(),
            include_deck=self.frm.includeDeck.isChecked(),
            include_notetype=self.frm.includeNotetype.isChecked(),
            include_guid=self.frm.includeGuid.isChecked(),
            legacy_support=self.frm.legacy_support.isChecked(),
            limit=limit,
        )

    def current_deck_id(self) -> DeckId | None:
        return (deck := self.current_deck()) and DeckId(deck.id) or None

    def current_deck(self) -> DeckNameId | None:
        if self.exporter.show_deck_list:
            if idx := self.frm.deck.currentIndex():
                return self.all_decks[idx - 1]
        return None

    def filename(self) -> str:
        if self.exporter.show_deck_list:
            deck_name = self.frm.deck.currentText()
            stem = re.sub('[\\\\/?<>:*|"^]', "_", deck_name)
        else:
            time_str = time.strftime("%Y-%m-%d@%H-%M-%S", time.localtime(time.time()))
            stem = f"{tr.exporting_collection()}-{time_str}"
        return f"{stem}.{self.exporter.extension}"


@dataclass
class ExportOptions:
    out_path: str
    include_scheduling: bool
    include_deck_configs: bool
    include_media: bool
    include_tags: bool
    include_html: bool
    include_deck: bool
    include_notetype: bool
    include_guid: bool
    legacy_support: bool
    limit: ExportLimit


class Exporter(ABC):
    extension: str
    show_deck_list = False
    show_include_scheduling = False
    show_include_deck_configs = False
    show_include_media = False
    show_include_tags = False
    show_include_html = False
    show_legacy_support = False
    show_include_deck = False
    show_include_notetype = False
    show_include_guid = False

    @abstractmethod
    def export(self, mw: aqt.main.AnkiQt, options: ExportOptions) -> None:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass


class ColpkgExporter(Exporter):
    extension = "colpkg"
    show_include_media = True
    show_legacy_support = True

    @staticmethod
    def name() -> str:
        return tr.exporting_anki_collection_package()

    def export(self, mw: aqt.main.AnkiQt, options: ExportOptions) -> None:
        options = gui_hooks.exporter_will_export(options, self)

        def on_success(_: None) -> None:
            mw.reopen()
            gui_hooks.exporter_did_export(options, self)
            tooltip(tr.exporting_collection_exported(), parent=mw)

        def on_failure(exception: Exception) -> None:
            mw.reopen()
            show_exception(parent=mw, exception=exception)

        gui_hooks.collection_will_temporarily_close(mw.col)
        QueryOp(
            parent=mw,
            op=lambda col: col.export_collection_package(
                options.out_path,
                include_media=options.include_media,
                legacy=options.legacy_support,
            ),
            success=on_success,
        ).with_backend_progress(export_progress_update).failure(
            on_failure
        ).run_in_background()


class ApkgExporter(Exporter):
    extension = "apkg"
    show_deck_list = True
    show_include_scheduling = True
    show_include_deck_configs = True
    show_include_media = True
    show_legacy_support = True

    @staticmethod
    def name() -> str:
        return tr.exporting_anki_deck_package()

    def export(self, mw: aqt.main.AnkiQt, options: ExportOptions) -> None:
        options = gui_hooks.exporter_will_export(options, self)

        def on_success(count: int) -> None:
            gui_hooks.exporter_did_export(options, self)
            tooltip(tr.exporting_note_exported(count=count), parent=mw)

        QueryOp(
            parent=mw,
            op=lambda col: col.export_anki_package(
                out_path=options.out_path,
                limit=options.limit,
                options=ExportAnkiPackageOptions(
                    with_scheduling=options.include_scheduling,
                    with_deck_configs=options.include_deck_configs,
                    with_media=options.include_media,
                    legacy=options.legacy_support,
                ),
            ),
            success=on_success,
        ).with_backend_progress(export_progress_update).run_in_background()


class NoteCsvExporter(Exporter):
    extension = "txt"
    show_deck_list = True
    show_include_html = True
    show_include_tags = True
    show_include_deck = True
    show_include_notetype = True
    show_include_guid = True

    @staticmethod
    def name() -> str:
        return tr.exporting_notes_in_plain_text()

    def export(self, mw: aqt.main.AnkiQt, options: ExportOptions) -> None:
        options = gui_hooks.exporter_will_export(options, self)

        def on_success(count: int) -> None:
            gui_hooks.exporter_did_export(options, self)
            tooltip(tr.exporting_note_exported(count=count), parent=mw)

        QueryOp(
            parent=mw,
            op=lambda col: col.export_note_csv(
                out_path=options.out_path,
                limit=options.limit,
                with_html=options.include_html,
                with_tags=options.include_tags,
                with_deck=options.include_deck,
                with_notetype=options.include_notetype,
                with_guid=options.include_guid,
            ),
            success=on_success,
        ).with_backend_progress(export_progress_update).run_in_background()


class CardCsvExporter(Exporter):
    extension = "txt"
    show_deck_list = True
    show_include_html = True

    @staticmethod
    def name() -> str:
        return tr.exporting_cards_in_plain_text()

    def export(self, mw: aqt.main.AnkiQt, options: ExportOptions) -> None:
        options = gui_hooks.exporter_will_export(options, self)

        def on_success(count: int) -> None:
            gui_hooks.exporter_did_export(options, self)
            tooltip(tr.exporting_card_exported(count=count), parent=mw)

        QueryOp(
            parent=mw,
            op=lambda col: col.export_card_csv(
                out_path=options.out_path,
                limit=options.limit,
                with_html=options.include_html,
            ),
            success=on_success,
        ).with_backend_progress(export_progress_update).run_in_background()


def export_progress_update(progress: Progress, update: ProgressUpdate) -> None:
    if not progress.HasField("exporting"):
        return
    update.label = progress.exporting
    if update.user_wants_abort:
        update.abort = True



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import aqt
from anki.buildinfo import buildhash
from anki.collection import CheckForUpdateResponse, Collection
from anki.utils import dev_mode, int_time, int_version, plat_desc
from aqt.operations import QueryOp
from aqt.package import (
    launcher_executable as _launcher_executable,
)
from aqt.package import (
    update_and_restart as _update_and_restart,
)
from aqt.qt import *
from aqt.utils import openLink, show_warning, showText, tr


def check_for_update() -> None:
    return

    from aqt import mw

    def do_check(_col: Collection) -> CheckForUpdateResponse:
        return mw.backend.check_for_update(
            version=int_version(),
            buildhash=buildhash,
            os=plat_desc(),
            install_id=mw.pm.meta["id"],
            last_message_id=max(0, mw.pm.meta["lastMsg"]),
        )

    def on_done(resp: CheckForUpdateResponse) -> None:
        # is clock off?
        if not dev_mode:
            diff = abs(resp.current_time - int_time())
            if diff > 300:
                diff_text = tr.qt_misc_second(count=diff)
                warn = (
                    tr.qt_misc_in_order_to_ensure_your_collection(val="%s") % diff_text
                )
                show_warning(
                    warn,
                    parent=mw,
                    textFormat=Qt.TextFormat.RichText,
                    callback=mw.app.closeAllWindows,
                )
                return
        # should we show a message?
        if msg := resp.message:
            showText(msg, parent=mw, type="html")
            mw.pm.meta["lastMsg"] = resp.last_message_id
        # has Anki been updated?
        if ver := resp.new_version:
            if mw.pm.meta.get("suppressUpdate", None) != ver:
                prompt_to_update(mw, ver)

    def on_fail(exc: Exception) -> None:
        print(f"update check failed: {exc}")

    QueryOp(parent=mw, op=do_check, success=on_done).failure(
        on_fail
    ).without_collection().run_in_background()


def prompt_to_update(mw: aqt.AnkiQt, ver: str) -> None:
    msg = (
        tr.qt_misc_anki_updatedanki_has_been_released(val=ver)
        + tr.qt_misc_would_you_like_to_download_it()
    )

    msgbox = QMessageBox(mw)
    msgbox.setStandardButtons(
        QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
    )
    msgbox.setIcon(QMessageBox.Icon.Information)
    msgbox.setText(msg)

    button = QPushButton(tr.qt_misc_ignore_this_update())
    msgbox.addButton(button, QMessageBox.ButtonRole.RejectRole)
    msgbox.setDefaultButton(QMessageBox.StandardButton.Yes)
    ret = msgbox.exec()

    if msgbox.clickedButton() == button:
        # ignore this update
        mw.pm.meta["suppressUpdate"] = ver
    elif ret == QMessageBox.StandardButton.Yes:
        if _launcher_executable():
            _update_and_restart()
        else:
            openLink(aqt.appWebsiteDownloadSection)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import functools
import os
import pathlib
import sys
import traceback
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Union

from anki._vendor import stringcase  # type: ignore

sys.modules["stringcase"] = stringcase

VariableTarget = tuple[Any, str]
DeprecatedAliasTarget = Union[Callable, VariableTarget]


def _target_to_string(target: DeprecatedAliasTarget | None) -> str:
    if target is None:
        return ""
    if name := getattr(target, "__name__", None):
        return name
    return target[1]  # type: ignore


def partial_path(full_path: str, components: int) -> str:
    path = pathlib.Path(full_path)
    return os.path.join(*path.parts[-components:])


def print_deprecation_warning(msg: str, frame: int = 1) -> None:
    # skip one frame to get to caller
    # then by default, skip one more frame as caller themself usually wants to
    # print their own caller
    path, linenum, _, _ = traceback.extract_stack(limit=frame + 2)[0]
    path = partial_path(path, components=3)
    print(f"{path}:{linenum}:{msg}")


def _print_warning(old: str, doc: str, frame: int = 1) -> None:
    return print_deprecation_warning(f"{old} is deprecated: {doc}", frame=frame + 1)


def _print_replacement_warning(old: str, new: str, frame: int = 1) -> None:
    doc = f"please use '{new}'" if new else "please implement your own"
    _print_warning(old, doc, frame=frame + 1)


def _get_remapped_and_replacement(
    mixin: DeprecatedNamesMixin | DeprecatedNamesMixinForModule, name: str
) -> tuple[str, str | None]:
    if some_tuple := mixin._deprecated_attributes.get(name):
        return some_tuple

    remapped = mixin._deprecated_aliases.get(name) or stringcase.snakecase(name)
    if remapped == name:
        raise AttributeError
    return (remapped, remapped)


class DeprecatedNamesMixin:
    "Expose instance methods/vars as camelCase for legacy callers."

    # deprecated name -> new name
    _deprecated_aliases: dict[str, str] = {}
    # deprecated name -> [new internal name, new name shown to user]
    _deprecated_attributes: dict[str, tuple[str, str | None]] = {}

    # TYPE_CHECKING check is required for https://github.com/python/mypy/issues/13319
    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            try:
                remapped, replacement = _get_remapped_and_replacement(self, name)
                out = getattr(self, remapped)
            except AttributeError:
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute '{name}'"
                ) from None

            _print_replacement_warning(name, replacement)
            return out

    @classmethod
    def register_deprecated_aliases(cls, **kwargs: DeprecatedAliasTarget) -> None:
        """Manually add aliases that are not a simple transform.

        Either pass in a method, or a tuple of (variable, "variable"). The
        latter is required because we want to ensure the provided arguments
        are valid symbols, and we can't get a variable's name easily.
        """
        cls._deprecated_aliases = {k: _target_to_string(v) for k, v in kwargs.items()}

    @classmethod
    def register_deprecated_attributes(
        cls,
        **kwargs: tuple[DeprecatedAliasTarget, DeprecatedAliasTarget | None],
    ) -> None:
        """Manually add deprecated attributes without exact substitutes.

        Pass a tuple of (alias, replacement), where alias is the attribute's new
        name (by convention: snakecase, prepended with '_legacy_'), and
        replacement is any callable to be used instead in new code or None.
        Also note the docstring of `register_deprecated_aliases`.

        E.g. given `def oldFunc(args): return new_func(additionalLogic(args))`,
        rename `oldFunc` to `_legacy_old_func` and call
        `register_deprecated_attributes(oldFunc=(_legacy_old_func, new_func))`.
        """
        cls._deprecated_attributes = {
            k: (_target_to_string(v[0]), _target_to_string(v[1]))
            for k, v in kwargs.items()
        }


class DeprecatedNamesMixinForModule:
    """Provides the functionality of DeprecatedNamesMixin for modules.

    It can be invoked like this:
    ```
        _deprecated_names = DeprecatedNamesMixinForModule(globals())
        _deprecated_names.register_deprecated_aliases(...
        _deprecated_names.register_deprecated_attributes(...

        if not TYPE_CHECKING:
            def __getattr__(name: str) -> Any:
                return _deprecated_names.__getattr__(name)
    ```
    See DeprecatedNamesMixin for more documentation.
    """

    def __init__(self, module_globals: dict[str, Any]) -> None:
        self.module_globals = module_globals
        self._deprecated_aliases: dict[str, str] = {}
        self._deprecated_attributes: dict[str, tuple[str, str | None]] = {}

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            try:
                remapped, replacement = _get_remapped_and_replacement(self, name)
                out = self.module_globals[remapped]
            except (AttributeError, KeyError):
                raise AttributeError(
                    f"Module '{self.module_globals['__name__']}' has no attribute '{name}'"
                ) from None

            # skip an additional frame as we are called from the module `__getattr__`
            _print_replacement_warning(name, replacement, frame=2)
            return out

    def register_deprecated_aliases(self, **kwargs: DeprecatedAliasTarget) -> None:
        self._deprecated_aliases = {k: _target_to_string(v) for k, v in kwargs.items()}

    def register_deprecated_attributes(
        self,
        **kwargs: tuple[DeprecatedAliasTarget, DeprecatedAliasTarget | None],
    ) -> None:
        self._deprecated_attributes = {
            k: (_target_to_string(v[0]), _target_to_string(v[1]))
            for k, v in kwargs.items()
        }


def deprecated(replaced_by: Callable | None = None, info: str = "") -> Callable:
    """Print a deprecation warning, telling users to use `replaced_by`, or show `doc`."""

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def decorated_func(*args: Any, **kwargs: Any) -> Any:
            if info:
                _print_warning(f"{func.__name__}()", info)
            else:
                _print_replacement_warning(func.__name__, replaced_by.__name__)

            return func(*args, **kwargs)

        return decorated_func

    return decorator


def deprecated_keywords(**replaced_keys: str) -> Callable:
    """Pass `oldKey="new_key"` to map the former to the latter, if passed to the
    decorated function as a key word, and print a deprecation warning.
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def decorated_func(*args: Any, **kwargs: Any) -> Any:
            updated_kwargs = {}
            for key, val in kwargs.items():
                if replacement := replaced_keys.get(key):
                    _print_replacement_warning(key, replacement)
                updated_kwargs[replacement or key] = val

            return func(*args, **updated_kwargs)

        return decorated_func

    return decorator



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

from anki._legacy import deprecated
from anki.cards import Card, CardId
from anki.consts import (
    CARD_TYPE_RELEARNING,
    QUEUE_TYPE_DAY_LEARN_RELEARN,
    QUEUE_TYPE_REV,
)
from anki.decks import DeckConfigDict, DeckId
from anki.notes import NoteId
from anki.scheduler.base import SchedulerBase, UnburyDeck
from anki.utils import from_json_bytes, ids2str


class SchedulerBaseWithLegacy(SchedulerBase):
    "Legacy aliases and helpers. These will go away in the future."

    def reschedCards(
        self, card_ids: list[CardId], min_interval: int, max_interval: int
    ) -> None:
        self.set_due_date(card_ids, f"{min_interval}-{max_interval}!")

    def buryNote(self, nid: NoteId) -> None:
        note = self.col.get_note(nid)
        self.bury_cards(note.card_ids())

    def unburyCards(self) -> None:
        print("please use unbury_cards() or unbury_deck() instead of unburyCards()")
        self.unbury_deck(self.col.decks.get_current_id())

    def unburyCardsForDeck(self, type: str = "all") -> None:
        print("please use unbury_deck() instead of unburyCardsForDeck()")
        if type == "all":
            mode = UnburyDeck.ALL
        elif type == "manual":
            mode = UnburyDeck.USER_ONLY
        else:  # elif type == "siblings":
            mode = UnburyDeck.SCHED_ONLY
        self.unbury_deck(self.col.decks.get_current_id(), mode)

    def finishedMsg(self) -> str:
        print("finishedMsg() is obsolete")
        return ""

    def _nextDueMsg(self) -> str:
        print("_nextDueMsg() is obsolete")
        return ""

    def rebuildDyn(self, did: DeckId | None = None) -> int | None:
        did = did or self.col.decks.selected()
        count = self.rebuild_filtered_deck(did).count or None
        if not count:
            return None
        # and change to our new deck
        self.col.decks.select(did)
        return count

    def emptyDyn(self, did: DeckId | None, lim: str | None = None) -> None:
        if lim is None:
            self.empty_filtered_deck(did)
            return

        queue = f"""
queue = (case when queue < 0 then queue
              when type in (1,{CARD_TYPE_RELEARNING}) then
  (case when (case when odue then odue else due end) > 1000000000 then 1 else
  {QUEUE_TYPE_DAY_LEARN_RELEARN} end)
else
  type
end)
"""
        self.col.db.execute(
            f"""
update cards set did = odid, {queue},
due = (case when odue>0 then odue else due end), odue = 0, odid = 0, usn = ? where {lim}""",
            self.col.usn(),
        )

    def remFromDyn(self, cids: list[CardId]) -> None:
        self.emptyDyn(None, f"id in {ids2str(cids)} and odid")

    # used by v2 scheduler and some add-ons
    def update_stats(
        self,
        deck_id: DeckId,
        new_delta: int = 0,
        review_delta: int = 0,
        milliseconds_delta: int = 0,
    ) -> None:
        self.col._backend.update_stats(
            deck_id=deck_id,
            new_delta=new_delta,
            review_delta=review_delta,
            millisecond_delta=milliseconds_delta,
        )

    def _updateStats(self, card: Card, type: str, cnt: int = 1) -> None:
        did = card.did
        if type == "new":
            self.update_stats(did, new_delta=cnt)
        elif type == "rev":
            self.update_stats(did, review_delta=cnt)
        elif type == "time":
            self.update_stats(did, milliseconds_delta=cnt)

    def deckDueTree(self) -> list:
        "List of (base name, did, rev, lrn, new, children)"
        print(
            "deckDueTree() is deprecated; use decks.deck_tree() for a tree without counts, or sched.deck_due_tree()"
        )
        return from_json_bytes(self.col._backend.deck_tree_legacy())[5]

    @deprecated(info="no longer used by Anki; will be removed in the future")
    def total_rev_for_current_deck(self) -> int:
        assert self.col.db
        return self.col.db.scalar(
            f"""
select count() from cards where id in (
select id from cards where did in %s and queue = {QUEUE_TYPE_REV} and due <= ? limit 9999)"""
            % self._deck_limit(),
            self.today,
        )

    def answerButtons(self, card: Card) -> int:
        return 4

    # legacy in v3 but used by unit tests; redefined in v2/v1

    def _cardConf(self, card: Card) -> DeckConfigDict:
        return self.col.decks.config_dict_for_deck_id(card.did)

    def _fuzzIvlRange(self, ivl: int) -> tuple[int, int]:
        return (ivl, ivl)

    # simple aliases
    unsuspendCards = SchedulerBase.unsuspend_cards
    buryCards = SchedulerBase.bury_cards
    suspendCards = SchedulerBase.suspend_cards
    forgetCards = SchedulerBase.schedule_cards_as_new



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import re
from concurrent.futures import Future
from typing import Any

import aqt
import aqt.forms
import aqt.main
from anki.cards import CardId
from anki.collection import EmptyCardsReport
from aqt import gui_hooks
from aqt.qt import QDialog, QDialogButtonBox, qconnect
from aqt.utils import disable_help_button, restoreGeom, saveGeom, tooltip, tr


def show_empty_cards(mw: aqt.main.AnkiQt) -> None:
    mw.progress.start()

    def on_done(fut: Future) -> None:
        mw.progress.finish()
        report: EmptyCardsReport = fut.result()
        if not report.notes:
            tooltip(tr.empty_cards_not_found())
            return
        diag = EmptyCardsDialog(mw, report)
        diag.show()

    mw.taskman.run_in_background(mw.col.get_empty_cards, on_done)


class EmptyCardsDialog(QDialog):
    silentlyClose = True

    def __init__(self, mw: aqt.main.AnkiQt, report: EmptyCardsReport) -> None:
        super().__init__(mw)
        self.mw = mw
        self.mw.garbage_collect_on_dialog_finish(self)
        self.report = report
        self.form = aqt.forms.emptycards.Ui_Dialog()
        self.form.setupUi(self)
        restoreGeom(self, "emptycards")
        self.setWindowTitle(tr.empty_cards_window_title())
        disable_help_button(self)
        self.form.keep_notes.setText(tr.empty_cards_preserve_notes_checkbox())
        self.form.webview.set_bridge_command(self._on_note_link_clicked, self)

        gui_hooks.empty_cards_will_show(self)

        # make the note ids clickable
        html = re.sub(
            r"\[anki:nid:(\d+)\]",
            "<a href=# onclick=\"pycmd('nid:\\1'); return false\">\\1</a>: ",
            report.report,
        )
        style = "<style>.allempty { color: red; }</style>"
        self.form.webview.stdHtml(style + html, context=self)

        def on_finished(code: Any) -> None:
            self.form.webview.cleanup()
            self.form.webview = None  # type: ignore
            saveGeom(self, "emptycards")

        qconnect(self.finished, on_finished)

        self._delete_button = self.form.buttonBox.addButton(
            tr.empty_cards_delete_button(), QDialogButtonBox.ButtonRole.ActionRole
        )
        assert self._delete_button is not None
        self._delete_button.setAutoDefault(False)
        qconnect(self._delete_button.clicked, self._on_delete)

    def _on_note_link_clicked(self, link: str) -> None:
        aqt.dialogs.open("Browser", self.mw, search=(link,))

    def _on_delete(self) -> None:
        self.mw.progress.start()

        def delete() -> int:
            return self._delete_cards(self.form.keep_notes.isChecked())

        def on_done(fut: Future) -> None:
            self.mw.progress.finish()
            try:
                count = fut.result()
            finally:
                self.close()
            tooltip(tr.empty_cards_deleted_count(cards=count))
            self.mw.reset()

        self.mw.taskman.run_in_background(delete, on_done)

    def _delete_cards(self, keep_notes: bool) -> int:
        to_delete: list[CardId] = []
        note: EmptyCardsReport.NoteWithEmptyCards
        for note in self.report.notes:
            if keep_notes and note.will_delete_note:
                # leave first card
                to_delete.extend([CardId(id) for id in note.card_ids[1:]])
            else:
                to_delete.extend([CardId(id) for id in note.card_ids])

        self.mw.col.remove_cards_and_orphaned_notes(to_delete)
        return len(to_delete)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Callable
from concurrent.futures._base import Future
from typing import Any, Generic, Protocol, TypeVar, Union

import aqt
import aqt.gui_hooks
import aqt.main
from anki.collection import (
    Collection,
    ImportLogWithChanges,
    OpChanges,
    OpChangesAfterUndo,
    OpChangesOnly,
    OpChangesWithCount,
    OpChangesWithId,
    Progress,
)
from aqt.errors import show_exception
from aqt.progress import ProgressUpdate
from aqt.qt import QWidget


class HasChangesProperty(Protocol):
    changes: OpChanges


# either an OpChanges object, or an object with .changes on it. This bound
# doesn't actually work for protobuf objects, so new protobuf objects will
# either need to be added here, or cast at call time
ResultWithChanges = TypeVar(
    "ResultWithChanges",
    bound=Union[
        OpChanges,
        OpChangesOnly,
        OpChangesWithCount,
        OpChangesWithId,
        OpChangesAfterUndo,
        ImportLogWithChanges,
        HasChangesProperty,
    ],
)


class CollectionOp(Generic[ResultWithChanges]):
    """Helper to perform a mutating DB operation on a background thread, and update UI.

    `op` should either return OpChanges, or an object with a 'changes'
    property. The changes will be passed to `operation_did_execute` so that
    the UI can decide whether it needs to update itself.

    - Shows progress popup for the duration of the op.
    - Ensures the browser doesn't try to redraw during the operation, which can lead
    to a frozen UI
    - Updates undo state at the end of the operation
    - Commits changes
    - Fires the `operation_(will|did)_reset` hooks
    - Fires the legacy `state_did_reset` hook

    Be careful not to call any UI routines in `op`, as that may crash Qt.
    This includes things select .selectedCards() in the browse screen.

    `success` will be called with the return value of op().

    If op() throws an exception, it will be shown in a popup, or
    passed to `failure` if it is provided.
    """

    _success: Callable[[ResultWithChanges], Any] | None = None
    _failure: Callable[[Exception], Any] | None = None
    _progress_update: Callable[[Progress, ProgressUpdate], None] | None = None

    def __init__(self, parent: QWidget, op: Callable[[Collection], ResultWithChanges]):
        self._parent = parent
        self._op = op

    def success(
        self, success: Callable[[ResultWithChanges], Any] | None
    ) -> CollectionOp[ResultWithChanges]:
        self._success = success
        return self

    def failure(
        self, failure: Callable[[Exception], Any] | None
    ) -> CollectionOp[ResultWithChanges]:
        self._failure = failure
        return self

    def with_backend_progress(
        self, progress_update: Callable[[Progress, ProgressUpdate], None] | None
    ) -> CollectionOp[ResultWithChanges]:
        self._progress_update = progress_update
        return self

    def run_in_background(self, *, initiator: object | None = None) -> None:
        from aqt import mw

        assert mw

        mw._increase_background_ops()

        def wrapped_op() -> ResultWithChanges:
            assert mw
            return self._op(mw.col)

        def wrapped_done(future: Future) -> None:
            assert mw
            mw._decrease_background_ops()
            # did something go wrong?
            if exception := future.exception():
                if isinstance(exception, Exception):
                    if self._failure:
                        self._failure(exception)
                    else:
                        show_exception(parent=self._parent, exception=exception)
                    return
                else:
                    # BaseException like SystemExit; rethrow it
                    future.result()

            result = future.result()
            try:
                if self._success:
                    self._success(result)
            finally:
                on_op_finished(mw, result, initiator)

        self._run(mw, wrapped_op, wrapped_done)

    def _run(
        self,
        mw: aqt.main.AnkiQt,
        op: Callable[[], ResultWithChanges],
        on_done: Callable[[Future], None],
    ) -> None:
        if self._progress_update:
            mw.taskman.with_backend_progress(
                op, self._progress_update, on_done=on_done, parent=self._parent
            )
        else:
            mw.taskman.with_progress(op, on_done, parent=self._parent)


def on_op_finished(
    mw: aqt.main.AnkiQt, result: ResultWithChanges, initiator: object | None
) -> None:
    mw.update_undo_actions()

    if isinstance(result, OpChanges):
        changes = result
    else:
        changes = result.changes  # type: ignore[union-attr]

    # fire new hook
    aqt.gui_hooks.operation_did_execute(changes, initiator)
    # fire legacy hook so old code notices changes
    if mw.col.op_made_changes(changes):
        aqt.gui_hooks.state_did_reset()


T = TypeVar("T")


class QueryOp(Generic[T]):
    """Helper to perform an operation on a background thread.

    QueryOp is primarily used for read-only requests (reading information
    from the database, fetching data from the network, etc), but can also
    be used for mutable requests outside of the collection undo system
    (eg adding/deleting files, calling a collection method that doesn't support
    undo, etc). For operations that support undo, use CollectionOp instead.

    - Optionally shows progress popup for the duration of the op.
    - Ensures the browser doesn't try to redraw during the operation, which can lead
    to a frozen UI

    Be careful not to call any UI routines in `op`, as that may crash Qt.
    This includes things like .selectedCards() in the browse screen.

    `success` will be called with the return value of op().

    If op() throws an exception, it will be shown in a popup, or
    passed to `failure` if it is provided.
    """

    _failure: Callable[[Exception], Any] | None = None
    _progress: bool | str = False
    _progress_update: Callable[[Progress, ProgressUpdate], None] | None = None

    def __init__(
        self,
        *,
        parent: QWidget,
        op: Callable[[Collection], T],
        success: Callable[[T], Any],
    ):
        self._parent = parent
        self._op = op
        self._success = success
        self._uses_collection = True

    def failure(self, failure: Callable[[Exception], Any] | None) -> QueryOp[T]:
        self._failure = failure
        return self

    def without_collection(self) -> QueryOp[T]:
        """Flag this QueryOp as not needing the collection.

        Operations that access the collection are serialized. If you're doing
        something like a series of network queries, and your operation does not
        access the collection, then you can call this to allow the requests to
        run in parallel."""
        self._uses_collection = False
        return self

    def with_progress(
        self,
        label: str | None = None,
    ) -> QueryOp[T]:
        "If label not provided, will default to 'Processing...'"
        self._progress = label or True
        return self

    def with_backend_progress(
        self, progress_update: Callable[[Progress, ProgressUpdate], None] | None
    ) -> QueryOp[T]:
        self._progress_update = progress_update
        return self

    def run_in_background(self) -> None:
        from aqt import mw

        assert mw

        mw._increase_background_ops()

        def wrapped_op() -> T:
            assert mw
            return self._op(mw.col)

        def wrapped_done(future: Future) -> None:
            assert mw

            mw._decrease_background_ops()
            # did something go wrong?
            if exception := future.exception():
                if isinstance(exception, Exception):
                    if self._failure:
                        self._failure(exception)
                    else:
                        show_exception(parent=self._parent, exception=exception)
                    return
                else:
                    # BaseException like SystemExit; rethrow it
                    future.result()

            self._success(future.result())

        self._run(mw, wrapped_op, wrapped_done)

    def _run(
        self,
        mw: aqt.main.AnkiQt,
        op: Callable[[], T],
        on_done: Callable[[Future], None],
    ) -> None:
        label = self._progress if isinstance(self._progress, str) else None
        if self._progress_update:
            mw.taskman.with_backend_progress(
                op,
                self._progress_update,
                on_done=on_done,
                start_label=label,
                parent=self._parent,
            )
        elif self._progress:
            mw.taskman.with_progress(op, on_done, label=label, parent=self._parent)
        else:
            mw.taskman.run_in_background(
                op, on_done, uses_collection=self._uses_collection
            )



# Form implementation generated from reading ui file 'qt/aqt/forms/main.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(667, 570)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QtCore.QSize(400, 0))
        MainWindow.setWindowTitle("Anki")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("icons:anki.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.centralwidget.setAutoFillBackground(True)
        self.centralwidget.setObjectName("centralwidget")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 667, 43))
        self.menubar.setObjectName("menubar")
        self.menuHelp = QtWidgets.QMenu(parent=self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuEdit = QtWidgets.QMenu(parent=self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuCol = QtWidgets.QMenu(parent=self.menubar)
        self.menuCol.setObjectName("menuCol")
        self.menuTools = QtWidgets.QMenu(parent=self.menubar)
        self.menuTools.setObjectName("menuTools")
        self.menuqt_accel_view = QtWidgets.QMenu(parent=self.menubar)
        self.menuqt_accel_view.setObjectName("menuqt_accel_view")
        MainWindow.setMenuBar(self.menubar)
        self.actionExit = QtGui.QAction(parent=MainWindow)
        self.actionExit.setShortcut("Ctrl+Q")
        self.actionExit.setObjectName("actionExit")
        self.actionPreferences = QtGui.QAction(parent=MainWindow)
        self.actionPreferences.setShortcut("Ctrl+P")
        self.actionPreferences.setMenuRole(QtGui.QAction.MenuRole.PreferencesRole)
        self.actionPreferences.setObjectName("actionPreferences")
        self.actionAbout = QtGui.QAction(parent=MainWindow)
        self.actionAbout.setMenuRole(QtGui.QAction.MenuRole.ApplicationSpecificRole)
        self.actionAbout.setObjectName("actionAbout")
        self.actionUndo = QtGui.QAction(parent=MainWindow)
        self.actionUndo.setEnabled(False)
        self.actionUndo.setShortcut("Ctrl+Z")
        self.actionUndo.setObjectName("actionUndo")
        self.actionCheckMediaDatabase = QtGui.QAction(parent=MainWindow)
        self.actionCheckMediaDatabase.setObjectName("actionCheckMediaDatabase")
        self.actionDonate = QtGui.QAction(parent=MainWindow)
        self.actionDonate.setObjectName("actionDonate")
        self.actionFullDatabaseCheck = QtGui.QAction(parent=MainWindow)
        self.actionFullDatabaseCheck.setObjectName("actionFullDatabaseCheck")
        self.actionDocumentation = QtGui.QAction(parent=MainWindow)
        self.actionDocumentation.setShortcut("F1")
        self.actionDocumentation.setObjectName("actionDocumentation")
        self.actionSwitchProfile = QtGui.QAction(parent=MainWindow)
        self.actionSwitchProfile.setShortcut("Ctrl+Shift+P")
        self.actionSwitchProfile.setObjectName("actionSwitchProfile")
        self.actionExport = QtGui.QAction(parent=MainWindow)
        self.actionExport.setShortcut("Ctrl+E")
        self.actionExport.setObjectName("actionExport")
        self.actionImport = QtGui.QAction(parent=MainWindow)
        self.actionImport.setShortcut("Ctrl+Shift+I")
        self.actionImport.setObjectName("actionImport")
        self.actionStudyDeck = QtGui.QAction(parent=MainWindow)
        self.actionStudyDeck.setShortcut("/")
        self.actionStudyDeck.setObjectName("actionStudyDeck")
        self.actionEmptyCards = QtGui.QAction(parent=MainWindow)
        self.actionEmptyCards.setObjectName("actionEmptyCards")
        self.actionCreateFiltered = QtGui.QAction(parent=MainWindow)
        self.actionCreateFiltered.setShortcut("F")
        self.actionCreateFiltered.setObjectName("actionCreateFiltered")
        self.actionNoteTypes = QtGui.QAction(parent=MainWindow)
        self.actionNoteTypes.setShortcut("Ctrl+Shift+N")
        self.actionNoteTypes.setObjectName("actionNoteTypes")
        self.actionAdd_ons = QtGui.QAction(parent=MainWindow)
        self.actionAdd_ons.setShortcut("Ctrl+Shift+A")
        self.actionAdd_ons.setObjectName("actionAdd_ons")
        self.actionRedo = QtGui.QAction(parent=MainWindow)
        self.actionRedo.setEnabled(False)
        self.actionRedo.setShortcut("Ctrl+Shift+Z")
        self.actionRedo.setObjectName("actionRedo")
        self.actionFullScreen = QtGui.QAction(parent=MainWindow)
        self.actionFullScreen.setObjectName("actionFullScreen")
        self.actionZoomIn = QtGui.QAction(parent=MainWindow)
        self.actionZoomIn.setObjectName("actionZoomIn")
        self.actionZoomOut = QtGui.QAction(parent=MainWindow)
        self.actionZoomOut.setObjectName("actionZoomOut")
        self.actionResetZoom = QtGui.QAction(parent=MainWindow)
        self.actionResetZoom.setShortcut("Ctrl+0")
        self.actionResetZoom.setObjectName("actionResetZoom")
        self.action_create_backup = QtGui.QAction(parent=MainWindow)
        self.action_create_backup.setObjectName("action_create_backup")
        self.action_open_backup = QtGui.QAction(parent=MainWindow)
        self.action_open_backup.setObjectName("action_open_backup")
        self.action_upgrade_downgrade = QtGui.QAction(parent=MainWindow)
        self.action_upgrade_downgrade.setObjectName("action_upgrade_downgrade")
        self.menuHelp.addAction(self.actionDocumentation)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionDonate)
        self.menuHelp.addAction(self.actionAbout)
        self.menuEdit.addAction(self.actionUndo)
        self.menuEdit.addAction(self.actionRedo)
        self.menuCol.addAction(self.actionSwitchProfile)
        self.menuCol.addSeparator()
        self.menuCol.addAction(self.actionImport)
        self.menuCol.addAction(self.actionExport)
        self.menuCol.addSeparator()
        self.menuCol.addAction(self.action_create_backup)
        self.menuCol.addAction(self.action_open_backup)
        self.menuCol.addSeparator()
        self.menuCol.addAction(self.actionExit)
        self.menuTools.addAction(self.actionStudyDeck)
        self.menuTools.addAction(self.actionCreateFiltered)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionFullDatabaseCheck)
        self.menuTools.addAction(self.actionCheckMediaDatabase)
        self.menuTools.addAction(self.actionEmptyCards)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionAdd_ons)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionNoteTypes)
        self.menuTools.addAction(self.action_upgrade_downgrade)
        self.menuTools.addAction(self.actionPreferences)
        self.menuqt_accel_view.addAction(self.actionFullScreen)
        self.menuqt_accel_view.addSeparator()
        self.menuqt_accel_view.addAction(self.actionZoomIn)
        self.menuqt_accel_view.addAction(self.actionZoomOut)
        self.menuqt_accel_view.addAction(self.actionResetZoom)
        self.menubar.addAction(self.menuCol.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuqt_accel_view.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        self.menuHelp.setTitle(tr.qt_accel_help())
        self.menuEdit.setTitle(tr.qt_accel_edit())
        self.menuCol.setTitle(tr.qt_accel_file())
        self.menuTools.setTitle(tr.qt_accel_tools())
        self.menuqt_accel_view.setTitle(tr.qt_accel_view())
        self.actionExit.setText(tr.qt_accel_exit())
        self.actionPreferences.setText(tr.qt_accel_preferences())
        self.actionPreferences.setStatusTip(tr.qt_misc_configure_interface_language_and_options())
        self.actionAbout.setText(tr.qt_accel_about())
        self.actionUndo.setText(tr.qt_accel_undo())
        self.actionCheckMediaDatabase.setText(tr.qt_accel_check_media())
        self.actionCheckMediaDatabase.setStatusTip(tr.qt_misc_check_the_files_in_the_media())
        self.actionDonate.setText(tr.qt_accel_support_anki())
        self.actionFullDatabaseCheck.setText(tr.qt_accel_check_database())
        self.actionDocumentation.setText(tr.qt_accel_guide())
        self.actionSwitchProfile.setText(tr.qt_accel_switch_profile())
        self.actionExport.setText(tr.qt_accel_export())
        self.actionImport.setText(tr.qt_accel_import())
        self.actionStudyDeck.setText(tr.qt_misc_study_deck())
        self.actionEmptyCards.setText(tr.qt_misc_empty_cards())
        self.actionCreateFiltered.setText(tr.qt_misc_create_filtered_deck())
        self.actionNoteTypes.setText(tr.qt_misc_manage_note_types())
        self.actionAdd_ons.setText(tr.qt_misc_addons())
        self.actionRedo.setText(tr.qt_accel_redo())
        self.actionFullScreen.setText(tr.qt_accel_full_screen())
        self.actionZoomIn.setText(tr.qt_accel_zoom_in())
        self.actionZoomOut.setText(tr.qt_accel_zoom_out())
        self.actionResetZoom.setText(tr.qt_accel_reset_zoom())
        self.action_create_backup.setText(tr.qt_accel_create_backup())
        self.action_open_backup.setText(tr.qt_accel_load_backup())
        self.action_upgrade_downgrade.setText(tr.qt_accel_upgrade_downgrade())


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/cards.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/cards.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import collection_pb2 as anki_dot_collection__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61nki/cards.proto\x12\nanki.cards\x1a\x15\x61nki/collection.proto\"\x15\n\x06\x43\x61rdId\x12\x0b\n\x03\x63id\x18\x01 \x01(\x03\"\x17\n\x07\x43\x61rdIds\x12\x0c\n\x04\x63ids\x18\x01 \x03(\x03\"\xd9\x04\n\x04\x43\x61rd\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0f\n\x07note_id\x18\x02 \x01(\x03\x12\x0f\n\x07\x64\x65\x63k_id\x18\x03 \x01(\x03\x12\x14\n\x0ctemplate_idx\x18\x04 \x01(\r\x12\x12\n\nmtime_secs\x18\x05 \x01(\x03\x12\x0b\n\x03usn\x18\x06 \x01(\x11\x12\r\n\x05\x63type\x18\x07 \x01(\r\x12\r\n\x05queue\x18\x08 \x01(\x11\x12\x0b\n\x03\x64ue\x18\t \x01(\x11\x12\x10\n\x08interval\x18\n \x01(\r\x12\x13\n\x0b\x65\x61se_factor\x18\x0b \x01(\r\x12\x0c\n\x04reps\x18\x0c \x01(\r\x12\x0e\n\x06lapses\x18\r \x01(\r\x12\x17\n\x0fremaining_steps\x18\x0e \x01(\r\x12\x14\n\x0coriginal_due\x18\x0f \x01(\x11\x12\x18\n\x10original_deck_id\x18\x10 \x01(\x03\x12\r\n\x05\x66lags\x18\x11 \x01(\r\x12\x1e\n\x11original_position\x18\x12 \x01(\rH\x00\x88\x01\x01\x12\x36\n\x0cmemory_state\x18\x14 \x01(\x0b\x32\x1b.anki.cards.FsrsMemoryStateH\x01\x88\x01\x01\x12\x1e\n\x11\x64\x65sired_retention\x18\x15 \x01(\x02H\x02\x88\x01\x01\x12\x12\n\x05\x64\x65\x63\x61y\x18\x16 \x01(\x02H\x03\x88\x01\x01\x12\"\n\x15last_review_time_secs\x18\x17 \x01(\x03H\x04\x88\x01\x01\x12\x13\n\x0b\x63ustom_data\x18\x13 \x01(\tB\x14\n\x12_original_positionB\x0f\n\r_memory_stateB\x14\n\x12_desired_retentionB\x08\n\x06_decayB\x18\n\x16_last_review_time_secs\"8\n\x0f\x46srsMemoryState\x12\x11\n\tstability\x18\x01 \x01(\x02\x12\x12\n\ndifficulty\x18\x02 \x01(\x02\"N\n\x12UpdateCardsRequest\x12\x1f\n\x05\x63\x61rds\x18\x01 \x03(\x0b\x32\x10.anki.cards.Card\x12\x17\n\x0fskip_undo_entry\x18\x02 \x01(\x08\"&\n\x12RemoveCardsRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\"3\n\x0eSetDeckRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x0f\n\x07\x64\x65\x63k_id\x18\x02 \x01(\x03\"0\n\x0eSetFlagRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x0c\n\x04\x66lag\x18\x02 \x01(\r2\xf6\x02\n\x0c\x43\x61rdsService\x12/\n\x07GetCard\x12\x12.anki.cards.CardId\x1a\x10.anki.cards.Card\x12I\n\x0bUpdateCards\x12\x1e.anki.cards.UpdateCardsRequest\x1a\x1a.anki.collection.OpChanges\x12R\n\x0bRemoveCards\x12\x1e.anki.cards.RemoveCardsRequest\x1a#.anki.collection.OpChangesWithCount\x12J\n\x07SetDeck\x12\x1a.anki.cards.SetDeckRequest\x1a#.anki.collection.OpChangesWithCount\x12J\n\x07SetFlag\x12\x1a.anki.cards.SetFlagRequest\x1a#.anki.collection.OpChangesWithCount2\x15\n\x13\x42\x61\x63kendCardsServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.cards_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_CARDID']._serialized_start=55
  _globals['_CARDID']._serialized_end=76
  _globals['_CARDIDS']._serialized_start=78
  _globals['_CARDIDS']._serialized_end=101
  _globals['_CARD']._serialized_start=104
  _globals['_CARD']._serialized_end=705
  _globals['_FSRSMEMORYSTATE']._serialized_start=707
  _globals['_FSRSMEMORYSTATE']._serialized_end=763
  _globals['_UPDATECARDSREQUEST']._serialized_start=765
  _globals['_UPDATECARDSREQUEST']._serialized_end=843
  _globals['_REMOVECARDSREQUEST']._serialized_start=845
  _globals['_REMOVECARDSREQUEST']._serialized_end=883
  _globals['_SETDECKREQUEST']._serialized_start=885
  _globals['_SETDECKREQUEST']._serialized_end=936
  _globals['_SETFLAGREQUEST']._serialized_start=938
  _globals['_SETFLAGREQUEST']._serialized_end=986
  _globals['_CARDSSERVICE']._serialized_start=989
  _globals['_CARDSSERVICE']._serialized_end=1363
  _globals['_BACKENDCARDSSERVICE']._serialized_start=1365
  _globals['_BACKENDCARDSSERVICE']._serialized_end=1386
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Callable

import aqt
import aqt.forms
import aqt.operations
from anki.collection import OpChangesWithId
from anki.decks import DeckId
from aqt import gui_hooks
from aqt.operations.deck import add_deck_dialog
from aqt.qt import *
from aqt.utils import (
    HelpPage,
    HelpPageArgument,
    add_close_shortcut,
    disable_help_button,
    openHelp,
    restoreGeom,
    saveGeom,
    shortcut,
    showInfo,
    tr,
)


class StudyDeck(QDialog):
    def __init__(
        self,
        mw: aqt.AnkiQt,
        names: Callable[[], list[str]] | None = None,
        accept: str | None = None,
        title: str | None = None,
        help: HelpPageArgument = HelpPage.KEYBOARD_SHORTCUTS,
        current: str | None = None,
        cancel: bool = True,
        parent: QWidget | None = None,
        dyn: bool = False,
        buttons: list[str | QPushButton] | None = None,
        geomKey: str = "default",
        callback: Callable[[StudyDeck], None] | None = None,
    ) -> None:
        super().__init__(parent)
        if not parent:
            mw.garbage_collect_on_dialog_finish(self)
        self.mw = mw
        self.form = aqt.forms.studydeck.Ui_Dialog()
        self.form.setupUi(self)
        self.form.filter.installEventFilter(self)
        gui_hooks.state_did_reset.append(self.onReset)
        self.geomKey = f"studyDeck-{geomKey}"
        restoreGeom(self, self.geomKey)
        add_close_shortcut(self)
        disable_help_button(self)
        if not cancel:
            self.form.buttonBox.removeButton(
                self.form.buttonBox.button(QDialogButtonBox.StandardButton.Cancel)
            )
        if buttons is not None:
            for button_or_label in buttons:
                self.form.buttonBox.addButton(
                    button_or_label, QDialogButtonBox.ButtonRole.ActionRole
                )
        else:
            b = QPushButton(tr.actions_add())
            b.setShortcut(QKeySequence("Ctrl+N"))
            b.setToolTip(shortcut(tr.decks_add_new_deck_ctrlandn()))
            self.form.buttonBox.addButton(b, QDialogButtonBox.ButtonRole.ActionRole)
            qconnect(b.clicked, self.onAddDeck)
        if title:
            self.setWindowTitle(title)
        if not names:
            names_ = [
                d.name
                for d in self.mw.col.decks.all_names_and_ids(
                    include_filtered=dyn, skip_empty_default=True
                )
            ]
            self.nameFunc = None
            self.origNames = names_
        else:
            self.nameFunc = names
            self.origNames = names()
        self.name: str | None = None
        self.form.buttonBox.addButton(
            accept or tr.decks_study(), QDialogButtonBox.ButtonRole.AcceptRole
        )
        self.setModal(True)
        qconnect(self.form.buttonBox.helpRequested, lambda: openHelp(help))
        qconnect(self.form.filter.textEdited, self.redraw)
        qconnect(self.form.list.itemDoubleClicked, self.accept)
        qconnect(self.finished, self.on_finished)
        self.form.filter.setFocus()
        self.show()
        # redraw after show so position at center correct
        self.redraw("", current)
        self.callback = callback
        if callback:
            self.show()
        else:
            self.exec()

    def eventFilter(self, obj: QObject | None, evt: QEvent | None) -> bool:
        if isinstance(evt, QKeyEvent) and evt.type() == QEvent.Type.KeyPress:
            new_row = current_row = self.form.list.currentRow()
            rows_count = self.form.list.count()
            key = evt.key()

            if key == Qt.Key.Key_Up:
                new_row = current_row - 1
            elif key == Qt.Key.Key_Down:
                new_row = current_row + 1
            elif (
                evt.modifiers() & Qt.KeyboardModifier.ControlModifier
                and Qt.Key.Key_1 <= key <= Qt.Key.Key_9
            ):
                row_index = key - Qt.Key.Key_1
                if row_index < rows_count:
                    new_row = row_index

            if rows_count:
                new_row %= rows_count  # don't let row index overflow/underflow
            if new_row != current_row:
                self.form.list.setCurrentRow(new_row)
                return True
        return False

    def redraw(self, filt: str, focus: str | None = None) -> None:
        self.filt = filt
        self.focus = focus
        self.names = [n for n in self.origNames if self._matches(n, filt)]
        l = self.form.list
        l.clear()
        l.addItems(self.names)
        if focus in self.names:
            idx = self.names.index(focus)
        else:
            idx = 0
        l.setCurrentRow(idx)
        l.scrollToItem(l.item(idx), QAbstractItemView.ScrollHint.PositionAtCenter)

    def _matches(self, name: str, filt: str) -> bool:
        name = name.lower()
        filt = filt.lower()
        if not filt:
            return True
        for word in filt.split(" "):
            if word not in name:
                return False
        return True

    def onReset(self) -> None:
        # model updated?
        if self.nameFunc:
            self.origNames = self.nameFunc()
        self.redraw(self.filt, self.focus)

    def accept(self) -> None:
        row = self.form.list.currentRow()
        if row < 0:
            showInfo(tr.decks_please_select_something())
            return
        self.name = self.names[self.form.list.currentRow()]
        self.accept_with_callback()

    def accept_with_callback(self) -> None:
        if self.callback:
            self.callback(self)
        super().accept()

    def onAddDeck(self) -> None:
        row = self.form.list.currentRow()
        if row < 0:
            default = self.form.filter.text()
        else:
            default = self.names[self.form.list.currentRow()]

        def success(out: OpChangesWithId) -> None:
            deck = self.mw.col.decks.get(DeckId(out.id))
            assert deck is not None
            self.name = deck["name"]
            self.accept_with_callback()

        if diag := add_deck_dialog(parent=self, default_text=default):
            diag.success(success).run_in_background()

    def on_finished(self) -> None:
        saveGeom(self, self.geomKey)
        gui_hooks.state_did_reset.remove(self.onReset)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/ankiweb.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/ankiweb.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x61nki/ankiweb.proto\x12\x0c\x61nki.ankiweb\"@\n\x13GetAddonInfoRequest\x12\x16\n\x0e\x63lient_version\x18\x01 \x01(\r\x12\x11\n\taddon_ids\x18\x02 \x03(\r\"=\n\x14GetAddonInfoResponse\x12%\n\x04info\x18\x01 \x03(\x0b\x32\x17.anki.ankiweb.AddonInfo\"S\n\tAddonInfo\x12\n\n\x02id\x18\x01 \x01(\r\x12\x10\n\x08modified\x18\x02 \x01(\x03\x12\x13\n\x0bmin_version\x18\x03 \x01(\r\x12\x13\n\x0bmax_version\x18\x04 \x01(\r\"t\n\x15\x43heckForUpdateRequest\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x11\n\tbuildhash\x18\x02 \x01(\t\x12\n\n\x02os\x18\x03 \x01(\t\x12\x12\n\ninstall_id\x18\x04 \x01(\x03\x12\x17\n\x0flast_message_id\x18\x05 \x01(\r\"\x93\x01\n\x16\x43heckForUpdateResponse\x12\x18\n\x0bnew_version\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x14\n\x0c\x63urrent_time\x18\x02 \x01(\x03\x12\x14\n\x07message\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x17\n\x0flast_message_id\x18\x04 \x01(\rB\x0e\n\x0c_new_versionB\n\n\x08_message2\x10\n\x0e\x41nkiwebService2\xcb\x01\n\x15\x42\x61\x63kendAnkiwebService\x12U\n\x0cGetAddonInfo\x12!.anki.ankiweb.GetAddonInfoRequest\x1a\".anki.ankiweb.GetAddonInfoResponse\x12[\n\x0e\x43heckForUpdate\x12#.anki.ankiweb.CheckForUpdateRequest\x1a$.anki.ankiweb.CheckForUpdateResponseB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.ankiweb_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_GETADDONINFOREQUEST']._serialized_start=36
  _globals['_GETADDONINFOREQUEST']._serialized_end=100
  _globals['_GETADDONINFORESPONSE']._serialized_start=102
  _globals['_GETADDONINFORESPONSE']._serialized_end=163
  _globals['_ADDONINFO']._serialized_start=165
  _globals['_ADDONINFO']._serialized_end=248
  _globals['_CHECKFORUPDATEREQUEST']._serialized_start=250
  _globals['_CHECKFORUPDATEREQUEST']._serialized_end=366
  _globals['_CHECKFORUPDATERESPONSE']._serialized_start=369
  _globals['_CHECKFORUPDATERESPONSE']._serialized_end=516
  _globals['_ANKIWEBSERVICE']._serialized_start=518
  _globals['_ANKIWEBSERVICE']._serialized_end=534
  _globals['_BACKENDANKIWEBSERVICE']._serialized_start=537
  _globals['_BACKENDANKIWEBSERVICE']._serialized_end=740
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/generic.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/generic.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x61nki/generic.proto\x12\x0c\x61nki.generic\"\x07\n\x05\x45mpty\"\x14\n\x05Int32\x12\x0b\n\x03val\x18\x01 \x01(\x11\"\x15\n\x06UInt32\x12\x0b\n\x03val\x18\x01 \x01(\r\"\x14\n\x05Int64\x12\x0b\n\x03val\x18\x01 \x01(\x03\"\x15\n\x06String\x12\x0b\n\x03val\x18\x01 \x01(\t\"\x14\n\x04Json\x12\x0c\n\x04json\x18\x01 \x01(\x0c\"\x13\n\x04\x42ool\x12\x0b\n\x03val\x18\x01 \x01(\x08\"\x1a\n\nStringList\x12\x0c\n\x04vals\x18\x01 \x03(\tB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.generic_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_EMPTY']._serialized_start=36
  _globals['_EMPTY']._serialized_end=43
  _globals['_INT32']._serialized_start=45
  _globals['_INT32']._serialized_end=65
  _globals['_UINT32']._serialized_start=67
  _globals['_UINT32']._serialized_end=88
  _globals['_INT64']._serialized_start=90
  _globals['_INT64']._serialized_end=110
  _globals['_STRING']._serialized_start=112
  _globals['_STRING']._serialized_end=133
  _globals['_JSON']._serialized_start=135
  _globals['_JSON']._serialized_end=155
  _globals['_BOOL']._serialized_start=157
  _globals['_BOOL']._serialized_end=176
  _globals['_STRINGLIST']._serialized_start=178
  _globals['_STRINGLIST']._serialized_end=204
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from concurrent.futures import Future

import aqt
import aqt.main
from aqt.qt import *
from aqt.utils import showText, tooltip


def on_progress(mw: aqt.main.AnkiQt) -> None:
    progress = mw.col.latest_progress()
    if not progress.HasField("database_check"):
        return
    dbprogress = progress.database_check
    mw.progress.update(
        process=False,
        label=dbprogress.stage,
        value=dbprogress.stage_current,
        max=dbprogress.stage_total,
    )


def check_db(mw: aqt.AnkiQt) -> None:
    def on_timer() -> None:
        on_progress(mw)

    timer = QTimer(mw)
    qconnect(timer.timeout, on_timer)
    timer.start(100)

    def do_check() -> tuple[str, bool]:
        mw.create_backup_now()
        return mw.col.fix_integrity()

    def on_future_done(fut: Future) -> None:
        timer.stop()
        ret, ok = fut.result()

        if not ok:
            showText(ret, parent=mw)
        else:
            tooltip(ret, parent=mw)

        # if an error has directed the user to check the database,
        # silently clean up any broken reset hooks which distract from
        # the underlying issue
        n = 0
        while n < 10:
            try:
                mw.reset()
                break
            except Exception as e:
                print("swallowed exception in reset hook:", e)
                n += 1
                continue

    mw.taskman.with_progress(do_check, on_future_done)



# Form implementation generated from reading ui file 'qt/aqt/forms/browser.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(750, 493)
        Dialog.setMinimumSize(QtCore.QSize(400, 400))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("icons:anki.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        Dialog.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(parent=Dialog)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setContentsMargins(0, 6, 6, 12)
        self.verticalLayout_3.setSpacing(12)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.splitter = QtWidgets.QSplitter(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(4)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy)
        self.splitter.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.splitter.setObjectName("splitter")
        self.widget = QtWidgets.QWidget(parent=self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setObjectName("widget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.widget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(0, 0, 0, 6)
        self.gridLayout.setHorizontalSpacing(12)
        self.gridLayout.setVerticalSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.searchEdit = QtWidgets.QComboBox(parent=self.widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(9)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.searchEdit.sizePolicy().hasHeightForWidth())
        self.searchEdit.setSizePolicy(sizePolicy)
        self.searchEdit.setEditable(True)
        self.searchEdit.setInsertPolicy(QtWidgets.QComboBox.InsertPolicy.NoInsert)
        self.searchEdit.setObjectName("searchEdit")
        self.gridLayout.addWidget(self.searchEdit, 0, 1, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.tableView = QtWidgets.QTableView(parent=self.widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(9)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.tableView.sizePolicy().hasHeightForWidth())
        self.tableView.setSizePolicy(sizePolicy)
        self.tableView.setMinimumSize(QtCore.QSize(0, 150))
        self.tableView.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.ActionsContextMenu)
        self.tableView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.tableView.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.tableView.setTabKeyNavigation(False)
        self.tableView.setAlternatingRowColors(True)
        self.tableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.tableView.setObjectName("tableView")
        self.tableView.horizontalHeader().setCascadingSectionResizes(False)
        self.tableView.horizontalHeader().setHighlightSections(False)
        self.tableView.horizontalHeader().setMinimumSectionSize(20)
        self.tableView.horizontalHeader().setSortIndicatorShown(True)
        self.verticalLayout_2.addWidget(self.tableView)
        self.verticalLayoutWidget = QtWidgets.QWidget(parent=self.splitter)
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(0, 1, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout2.setSpacing(0)
        self.horizontalLayout2.setObjectName("horizontalLayout2")
        self.fieldsArea = QtWidgets.QWidget(parent=self.verticalLayoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.fieldsArea.sizePolicy().hasHeightForWidth())
        self.fieldsArea.setSizePolicy(sizePolicy)
        self.fieldsArea.setMinimumSize(QtCore.QSize(50, 200))
        self.fieldsArea.setObjectName("fieldsArea")
        self.horizontalLayout2.addWidget(self.fieldsArea)
        self.verticalLayout.addLayout(self.horizontalLayout2)
        self.verticalLayout_3.addWidget(self.splitter)
        Dialog.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=Dialog)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 750, 23))
        self.menubar.setObjectName("menubar")
        self.menuEdit = QtWidgets.QMenu(parent=self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuJump = QtWidgets.QMenu(parent=self.menubar)
        self.menuJump.setObjectName("menuJump")
        self.menu_Help = QtWidgets.QMenu(parent=self.menubar)
        self.menu_Help.setObjectName("menu_Help")
        self.menu_Cards = QtWidgets.QMenu(parent=self.menubar)
        self.menu_Cards.setObjectName("menu_Cards")
        self.menuFlag = QtWidgets.QMenu(parent=self.menu_Cards)
        self.menuFlag.setObjectName("menuFlag")
        self.menu_Notes = QtWidgets.QMenu(parent=self.menubar)
        self.menu_Notes.setObjectName("menu_Notes")
        self.menuqt_accel_view = QtWidgets.QMenu(parent=self.menubar)
        self.menuqt_accel_view.setObjectName("menuqt_accel_view")
        self.menuLayout = QtWidgets.QMenu(parent=self.menuqt_accel_view)
        self.menuLayout.setObjectName("menuLayout")
        Dialog.setMenuBar(self.menubar)
        self.actionSelectAll = QtGui.QAction(parent=Dialog)
        self.actionSelectAll.setShortcut("Ctrl+Alt+A")
        self.actionSelectAll.setObjectName("actionSelectAll")
        self.actionUndo = QtGui.QAction(parent=Dialog)
        self.actionUndo.setShortcut("Ctrl+Z")
        self.actionUndo.setObjectName("actionUndo")
        self.actionInvertSelection = QtGui.QAction(parent=Dialog)
        self.actionInvertSelection.setShortcut("Ctrl+Alt+S")
        self.actionInvertSelection.setObjectName("actionInvertSelection")
        self.actionFind = QtGui.QAction(parent=Dialog)
        self.actionFind.setShortcut("Ctrl+F")
        self.actionFind.setObjectName("actionFind")
        self.actionNote = QtGui.QAction(parent=Dialog)
        self.actionNote.setShortcut("Ctrl+Shift+N")
        self.actionNote.setObjectName("actionNote")
        self.actionNextCard = QtGui.QAction(parent=Dialog)
        self.actionNextCard.setShortcut("Ctrl+N")
        self.actionNextCard.setObjectName("actionNextCard")
        self.actionPreviousCard = QtGui.QAction(parent=Dialog)
        self.actionPreviousCard.setShortcut("Ctrl+P")
        self.actionPreviousCard.setObjectName("actionPreviousCard")
        self.actionGuide = QtGui.QAction(parent=Dialog)
        self.actionGuide.setShortcut("F1")
        self.actionGuide.setObjectName("actionGuide")
        self.actionChangeModel = QtGui.QAction(parent=Dialog)
        self.actionChangeModel.setShortcut("Ctrl+Shift+M")
        self.actionChangeModel.setObjectName("actionChangeModel")
        self.actionSelectNotes = QtGui.QAction(parent=Dialog)
        self.actionSelectNotes.setObjectName("actionSelectNotes")
        self.actionFindReplace = QtGui.QAction(parent=Dialog)
        self.actionFindReplace.setShortcut("Ctrl+Alt+F")
        self.actionFindReplace.setObjectName("actionFindReplace")
        self.actionSidebarFilter = QtGui.QAction(parent=Dialog)
        self.actionSidebarFilter.setShortcut("Ctrl+Shift+F")
        self.actionSidebarFilter.setObjectName("actionSidebarFilter")
        self.actionCardList = QtGui.QAction(parent=Dialog)
        self.actionCardList.setShortcut("Ctrl+Shift+L")
        self.actionCardList.setObjectName("actionCardList")
        self.actionFindDuplicates = QtGui.QAction(parent=Dialog)
        self.actionFindDuplicates.setObjectName("actionFindDuplicates")
        self.actionReposition = QtGui.QAction(parent=Dialog)
        self.actionReposition.setShortcut("Ctrl+Shift+S")
        self.actionReposition.setObjectName("actionReposition")
        self.actionFirstCard = QtGui.QAction(parent=Dialog)
        self.actionFirstCard.setShortcut("Home")
        self.actionFirstCard.setObjectName("actionFirstCard")
        self.actionLastCard = QtGui.QAction(parent=Dialog)
        self.actionLastCard.setShortcut("End")
        self.actionLastCard.setObjectName("actionLastCard")
        self.actionClose = QtGui.QAction(parent=Dialog)
        self.actionClose.setShortcut("Ctrl+W")
        self.actionClose.setObjectName("actionClose")
        self.action_Info = QtGui.QAction(parent=Dialog)
        self.action_Info.setShortcut("Ctrl+Shift+I")
        self.action_Info.setObjectName("action_Info")
        self.actionAdd_Tags = QtGui.QAction(parent=Dialog)
        self.actionAdd_Tags.setShortcut("Ctrl+Shift+A")
        self.actionAdd_Tags.setObjectName("actionAdd_Tags")
        self.actionRemove_Tags = QtGui.QAction(parent=Dialog)
        self.actionRemove_Tags.setShortcut("Ctrl+Alt+Shift+A")
        self.actionRemove_Tags.setObjectName("actionRemove_Tags")
        self.actionToggle_Suspend = QtGui.QAction(parent=Dialog)
        self.actionToggle_Suspend.setCheckable(True)
        self.actionToggle_Suspend.setShortcut("Ctrl+J")
        self.actionToggle_Suspend.setObjectName("actionToggle_Suspend")
        self.actionDelete = QtGui.QAction(parent=Dialog)
        self.actionDelete.setShortcut("Ctrl+Del")
        self.actionDelete.setObjectName("actionDelete")
        self.actionAdd = QtGui.QAction(parent=Dialog)
        self.actionAdd.setShortcut("Ctrl+E")
        self.actionAdd.setObjectName("actionAdd")
        self.actionChange_Deck = QtGui.QAction(parent=Dialog)
        self.actionChange_Deck.setShortcut("Ctrl+D")
        self.actionChange_Deck.setObjectName("actionChange_Deck")
        self.actionRed_Flag = QtGui.QAction(parent=Dialog)
        self.actionRed_Flag.setCheckable(True)
        self.actionRed_Flag.setShortcut("Ctrl+1")
        self.actionRed_Flag.setObjectName("actionRed_Flag")
        self.actionOrange_Flag = QtGui.QAction(parent=Dialog)
        self.actionOrange_Flag.setCheckable(True)
        self.actionOrange_Flag.setShortcut("Ctrl+2")
        self.actionOrange_Flag.setObjectName("actionOrange_Flag")
        self.actionGreen_Flag = QtGui.QAction(parent=Dialog)
        self.actionGreen_Flag.setCheckable(True)
        self.actionGreen_Flag.setShortcut("Ctrl+3")
        self.actionGreen_Flag.setObjectName("actionGreen_Flag")
        self.actionBlue_Flag = QtGui.QAction(parent=Dialog)
        self.actionBlue_Flag.setCheckable(True)
        self.actionBlue_Flag.setShortcut("Ctrl+4")
        self.actionBlue_Flag.setObjectName("actionBlue_Flag")
        self.actionSidebar = QtGui.QAction(parent=Dialog)
        self.actionSidebar.setShortcut("Ctrl+Shift+R")
        self.actionSidebar.setObjectName("actionSidebar")
        self.actionClear_Unused_Tags = QtGui.QAction(parent=Dialog)
        self.actionClear_Unused_Tags.setObjectName("actionClear_Unused_Tags")
        self.actionManage_Note_Types = QtGui.QAction(parent=Dialog)
        self.actionManage_Note_Types.setObjectName("actionManage_Note_Types")
        self.actionToggle_Mark = QtGui.QAction(parent=Dialog)
        self.actionToggle_Mark.setCheckable(True)
        self.actionToggle_Mark.setShortcut("Ctrl+K")
        self.actionToggle_Mark.setObjectName("actionToggle_Mark")
        self.actionExport = QtGui.QAction(parent=Dialog)
        self.actionExport.setShortcut("Ctrl+Shift+E")
        self.actionExport.setObjectName("actionExport")
        self.actionCreateFilteredDeck = QtGui.QAction(parent=Dialog)
        self.actionCreateFilteredDeck.setShortcut("Ctrl+G")
        self.actionCreateFilteredDeck.setObjectName("actionCreateFilteredDeck")
        self.action_set_due_date = QtGui.QAction(parent=Dialog)
        self.action_set_due_date.setShortcut("Ctrl+Shift+D")
        self.action_set_due_date.setObjectName("action_set_due_date")
        self.action_grade_now = QtGui.QAction(parent=Dialog)
        self.action_grade_now.setShortcut("Ctrl+Shift+G")
        self.action_grade_now.setObjectName("action_grade_now")
        self.action_forget = QtGui.QAction(parent=Dialog)
        self.action_forget.setShortcut("Ctrl+Alt+N")
        self.action_forget.setObjectName("action_forget")
        self.action_toggle_mode = QtGui.QAction(parent=Dialog)
        self.action_toggle_mode.setShortcut("Ctrl+Alt+T")
        self.action_toggle_mode.setObjectName("action_toggle_mode")
        self.actionRedo = QtGui.QAction(parent=Dialog)
        self.actionRedo.setShortcut("Ctrl+Shift+Z")
        self.actionRedo.setObjectName("actionRedo")
        self.actionPink_Flag = QtGui.QAction(parent=Dialog)
        self.actionPink_Flag.setCheckable(True)
        self.actionPink_Flag.setShortcut("Ctrl+5")
        self.actionPink_Flag.setObjectName("actionPink_Flag")
        self.actionTurquoise_Flag = QtGui.QAction(parent=Dialog)
        self.actionTurquoise_Flag.setCheckable(True)
        self.actionTurquoise_Flag.setShortcut("Ctrl+6")
        self.actionTurquoise_Flag.setObjectName("actionTurquoise_Flag")
        self.actionPurple_Flag = QtGui.QAction(parent=Dialog)
        self.actionPurple_Flag.setCheckable(True)
        self.actionPurple_Flag.setShortcut("Ctrl+7")
        self.actionPurple_Flag.setObjectName("actionPurple_Flag")
        self.actionCopy = QtGui.QAction(parent=Dialog)
        self.actionCopy.setShortcut("Ctrl+Alt+E")
        self.actionCopy.setObjectName("actionCopy")
        self.actionFullScreen = QtGui.QAction(parent=Dialog)
        self.actionFullScreen.setObjectName("actionFullScreen")
        self.actionToggleSidebar = QtGui.QAction(parent=Dialog)
        self.actionToggleSidebar.setObjectName("actionToggleSidebar")
        self.actionZoomIn = QtGui.QAction(parent=Dialog)
        self.actionZoomIn.setObjectName("actionZoomIn")
        self.actionZoomOut = QtGui.QAction(parent=Dialog)
        self.actionZoomOut.setObjectName("actionZoomOut")
        self.actionResetZoom = QtGui.QAction(parent=Dialog)
        self.actionResetZoom.setShortcut("Ctrl+0")
        self.actionResetZoom.setObjectName("actionResetZoom")
        self.actionLayoutAuto = QtGui.QAction(parent=Dialog)
        self.actionLayoutAuto.setCheckable(True)
        self.actionLayoutAuto.setObjectName("actionLayoutAuto")
        self.actionLayoutVertical = QtGui.QAction(parent=Dialog)
        self.actionLayoutVertical.setCheckable(True)
        self.actionLayoutVertical.setObjectName("actionLayoutVertical")
        self.actionLayoutHorizontal = QtGui.QAction(parent=Dialog)
        self.actionLayoutHorizontal.setCheckable(True)
        self.actionLayoutHorizontal.setObjectName("actionLayoutHorizontal")
        self.actionbrowsing_toggle_showing_cards_notes = QtGui.QAction(parent=Dialog)
        self.actionbrowsing_toggle_showing_cards_notes.setObjectName("actionbrowsing_toggle_showing_cards_notes")
        self.action_toggle_bury = QtGui.QAction(parent=Dialog)
        self.action_toggle_bury.setCheckable(True)
        self.action_toggle_bury.setShortcut("Ctrl+Shift+J")
        self.action_toggle_bury.setObjectName("action_toggle_bury")
        self.menuEdit.addAction(self.actionUndo)
        self.menuEdit.addAction(self.actionRedo)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionSelectAll)
        self.menuEdit.addAction(self.actionSelectNotes)
        self.menuEdit.addAction(self.actionInvertSelection)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionClose)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionCreateFilteredDeck)
        self.menuJump.addAction(self.actionFind)
        self.menuJump.addAction(self.actionSidebarFilter)
        self.menuJump.addAction(self.actionSidebar)
        self.menuJump.addAction(self.actionNote)
        self.menuJump.addAction(self.actionCardList)
        self.menuJump.addSeparator()
        self.menuJump.addAction(self.actionFirstCard)
        self.menuJump.addAction(self.actionPreviousCard)
        self.menuJump.addAction(self.actionNextCard)
        self.menuJump.addAction(self.actionLastCard)
        self.menu_Help.addAction(self.actionGuide)
        self.menuFlag.addAction(self.actionRed_Flag)
        self.menuFlag.addAction(self.actionOrange_Flag)
        self.menuFlag.addAction(self.actionGreen_Flag)
        self.menuFlag.addAction(self.actionBlue_Flag)
        self.menuFlag.addAction(self.actionPink_Flag)
        self.menuFlag.addAction(self.actionTurquoise_Flag)
        self.menuFlag.addAction(self.actionPurple_Flag)
        self.menu_Cards.addAction(self.actionChange_Deck)
        self.menu_Cards.addSeparator()
        self.menu_Cards.addAction(self.action_set_due_date)
        self.menu_Cards.addAction(self.action_grade_now)
        self.menu_Cards.addAction(self.action_forget)
        self.menu_Cards.addAction(self.actionReposition)
        self.menu_Cards.addSeparator()
        self.menu_Cards.addAction(self.actionToggle_Suspend)
        self.menu_Cards.addAction(self.action_toggle_bury)
        self.menu_Cards.addSeparator()
        self.menu_Cards.addAction(self.menuFlag.menuAction())
        self.menu_Cards.addSeparator()
        self.menu_Cards.addAction(self.action_Info)
        self.menu_Notes.addAction(self.actionAdd)
        self.menu_Notes.addAction(self.actionCopy)
        self.menu_Notes.addAction(self.actionExport)
        self.menu_Notes.addSeparator()
        self.menu_Notes.addAction(self.actionAdd_Tags)
        self.menu_Notes.addAction(self.actionRemove_Tags)
        self.menu_Notes.addAction(self.actionClear_Unused_Tags)
        self.menu_Notes.addAction(self.actionToggle_Mark)
        self.menu_Notes.addSeparator()
        self.menu_Notes.addAction(self.actionChangeModel)
        self.menu_Notes.addSeparator()
        self.menu_Notes.addAction(self.actionFindDuplicates)
        self.menu_Notes.addAction(self.actionFindReplace)
        self.menu_Notes.addSeparator()
        self.menu_Notes.addAction(self.actionManage_Note_Types)
        self.menu_Notes.addSeparator()
        self.menu_Notes.addAction(self.actionDelete)
        self.menuLayout.addAction(self.actionLayoutAuto)
        self.menuLayout.addSeparator()
        self.menuLayout.addAction(self.actionLayoutVertical)
        self.menuLayout.addAction(self.actionLayoutHorizontal)
        self.menuqt_accel_view.addAction(self.action_toggle_mode)
        self.menuqt_accel_view.addSeparator()
        self.menuqt_accel_view.addAction(self.actionFullScreen)
        self.menuqt_accel_view.addSeparator()
        self.menuqt_accel_view.addAction(self.actionToggleSidebar)
        self.menuqt_accel_view.addSeparator()
        self.menuqt_accel_view.addAction(self.actionZoomIn)
        self.menuqt_accel_view.addAction(self.actionZoomOut)
        self.menuqt_accel_view.addAction(self.actionResetZoom)
        self.menuqt_accel_view.addSeparator()
        self.menuqt_accel_view.addAction(self.menuLayout.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuqt_accel_view.menuAction())
        self.menubar.addAction(self.menu_Notes.menuAction())
        self.menubar.addAction(self.menu_Cards.menuAction())
        self.menubar.addAction(self.menuJump.menuAction())
        self.menubar.addAction(self.menu_Help.menuAction())

        self.retranslateUi(Dialog)
        self.actionSelectAll.triggered.connect(self.tableView.selectAll) # type: ignore  # type: ignore
        self.actionClose.triggered.connect(Dialog._handle_close) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        self.menuEdit.setTitle(tr.qt_accel_edit())
        self.menuJump.setTitle(tr.qt_accel_go())
        self.menu_Help.setTitle(tr.qt_accel_help())
        self.menu_Cards.setTitle(tr.qt_accel_cards())
        self.menuFlag.setTitle(tr.browsing_flag())
        self.menu_Notes.setTitle(tr.qt_accel_notes())
        self.menuqt_accel_view.setTitle(tr.qt_accel_view())
        self.menuLayout.setTitle(tr.qt_accel_layout())
        self.actionSelectAll.setText(tr.qt_accel_select_all())
        self.actionUndo.setText(tr.qt_accel_undo())
        self.actionInvertSelection.setText(tr.qt_accel_invert_selection())
        self.actionFind.setText(tr.qt_accel_find())
        self.actionNote.setText(tr.qt_accel_note())
        self.actionNextCard.setText(tr.qt_accel_next_card())
        self.actionPreviousCard.setText(tr.qt_accel_previous_card())
        self.actionGuide.setText(tr.qt_accel_guide())
        self.actionChangeModel.setText(tr.browsing_change_note_type2())
        self.actionSelectNotes.setText(tr.qt_accel_select_notes())
        self.actionFindReplace.setText(tr.qt_accel_find_and_replace())
        self.actionSidebarFilter.setText(tr.qt_accel_filter())
        self.actionCardList.setText(tr.browsing_card_list())
        self.actionFindDuplicates.setText(tr.qt_accel_find_duplicates())
        self.actionReposition.setText(tr.browsing_reposition())
        self.actionFirstCard.setText(tr.browsing_first_card())
        self.actionLastCard.setText(tr.browsing_last_card())
        self.actionClose.setText(tr.actions_close())
        self.action_Info.setText(tr.qt_accel_info())
        self.actionAdd_Tags.setText(tr.browsing_add_tags2())
        self.actionRemove_Tags.setText(tr.browsing_remove_tags())
        self.actionToggle_Suspend.setText(tr.browsing_toggle_suspend())
        self.actionDelete.setText(tr.actions_delete())
        self.actionAdd.setText(tr.browsing_add_notes())
        self.actionChange_Deck.setText(tr.browsing_change_deck2())
        self.actionRed_Flag.setText(tr.actions_flag_red())
        self.actionOrange_Flag.setText(tr.actions_flag_orange())
        self.actionGreen_Flag.setText(tr.actions_flag_green())
        self.actionBlue_Flag.setText(tr.actions_flag_blue())
        self.actionSidebar.setText(tr.browsing_sidebar())
        self.actionClear_Unused_Tags.setText(tr.browsing_clear_unused_tags())
        self.actionManage_Note_Types.setText(tr.browsing_manage_note_types())
        self.actionToggle_Mark.setText(tr.browsing_toggle_mark())
        self.actionExport.setText(tr.qt_accel_export_notes())
        self.actionCreateFilteredDeck.setText(tr.qt_misc_create_filtered_deck())
        self.action_set_due_date.setText(tr.qt_accel_set_due_date())
        self.action_grade_now.setText(tr.actions_grade_now())
        self.action_forget.setText(tr.qt_accel_forget())
        self.action_toggle_mode.setText(tr.browsing_toggle_showing_cards_notes())
        self.actionRedo.setText(tr.qt_accel_redo())
        self.actionPink_Flag.setText(tr.actions_flag_pink())
        self.actionTurquoise_Flag.setText(tr.actions_flag_turquoise())
        self.actionPurple_Flag.setText(tr.actions_flag_purple())
        self.actionCopy.setText(tr.actions_create_copy())
        self.actionFullScreen.setText(tr.qt_accel_full_screen())
        self.actionToggleSidebar.setText(tr.qt_accel_toggle_sidebar())
        self.actionZoomIn.setText(tr.qt_accel_zoom_editor_in())
        self.actionZoomOut.setText(tr.qt_accel_zoom_editor_out())
        self.actionResetZoom.setText(tr.qt_accel_reset_zoom())
        self.actionLayoutAuto.setText(tr.qt_accel_layout_auto())
        self.actionLayoutVertical.setText(tr.qt_accel_layout_vertical())
        self.actionLayoutHorizontal.setText(tr.qt_accel_layout_horizontal())
        self.actionbrowsing_toggle_showing_cards_notes.setText(tr.browsing_toggle_showing_cards_notes())
        self.action_toggle_bury.setText(tr.browsing_toggle_bury())


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import os
import os.path
import platform
import re
import subprocess
import sys
import time
import traceback
import wave
from abc import ABC, abstractmethod
from collections.abc import Callable
from concurrent.futures import Future
from operator import itemgetter
from pathlib import Path
from typing import Any, cast

from markdown import markdown

import aqt
import aqt.mpv
import aqt.qt
from anki.cards import Card
from anki.sound import AV_REF_RE, AVTag, SoundOrVideoTag
from anki.utils import is_lin, is_mac, is_win, namedtmp
from aqt import gui_hooks
from aqt._macos_helper import macos_helper
from aqt.mpv import MPV, MPVBase, MPVCommandError
from aqt.qt import *
from aqt.taskman import TaskManager
from aqt.theme import theme_manager
from aqt.utils import (
    disable_help_button,
    restoreGeom,
    saveGeom,
    showWarning,
    startup_info,
    tooltip,
    tr,
)

# AV player protocol
##########################################################################

OnDoneCallback = Callable[[], None]


class Player(ABC):
    @abstractmethod
    def play(self, tag: AVTag, on_done: OnDoneCallback) -> None:
        """Play a file.

        When reimplementing, make sure to call
        gui_hooks.av_player_did_begin_playing(self, tag)
        on the main thread after playback begins.
        """

    @abstractmethod
    def rank_for_tag(self, tag: AVTag) -> int | None:
        """How suited this player is to playing tag.

        AVPlayer will choose the player that returns the highest rank
        for a given tag.

        If None, this player can not play the tag.
        """

    def stop(self) -> None:
        """Optional.

        If implemented, the player must not call on_done() when the audio is stopped."""

    def seek_relative(self, secs: int) -> None:
        "Jump forward or back by secs. Optional."

    def toggle_pause(self) -> None:
        "Optional."

    def shutdown(self) -> None:
        "Do any cleanup required at program termination. Optional."


AUDIO_EXTENSIONS = {
    "3gp",
    "flac",
    "m4a",
    "mp3",
    "oga",
    "ogg",
    "opus",
    "spx",
    "wav",
}


def is_audio_file(fname: str) -> bool:
    ext = fname.split(".")[-1].lower()
    return ext in AUDIO_EXTENSIONS


class SoundOrVideoPlayer(Player):
    default_rank = 0

    def rank_for_tag(self, tag: AVTag) -> int | None:
        if isinstance(tag, SoundOrVideoTag):
            return self.default_rank
        else:
            return None


class SoundPlayer(Player):
    default_rank = 0

    def rank_for_tag(self, tag: AVTag) -> int | None:
        if isinstance(tag, SoundOrVideoTag) and is_audio_file(tag.filename):
            return self.default_rank
        else:
            return None


class VideoPlayer(Player):
    default_rank = 0

    def rank_for_tag(self, tag: AVTag) -> int | None:
        if isinstance(tag, SoundOrVideoTag) and not is_audio_file(tag.filename):
            return self.default_rank
        else:
            return None


# Main playing interface
##########################################################################


class AVPlayer:
    players: list[Player] = []
    # when a new batch of audio is played, should the currently playing
    # audio be stopped?
    interrupt_current_audio = True
    # caller key for the current playback (optional)
    current_caller: Any = None
    # whether the last call to play_file_with_caller interrupted another
    current_caller_interrupted = False

    def __init__(self) -> None:
        self._enqueued: list[AVTag] = []
        self.current_player: Player | None = None

    def play_tags(self, tags: list[AVTag]) -> None:
        """Clear the existing queue, then start playing provided tags."""
        self.clear_queue_and_maybe_interrupt()
        self._enqueued = tags[:]
        self._play_next_if_idle()

    def append_tags(self, tags: list[AVTag]) -> None:
        """Append provided tags to the queue, then start playing them if the current player is idle."""
        self._enqueued.extend(tags)
        self._play_next_if_idle()

    def queue_is_empty(self) -> bool:
        return not bool(self._enqueued)

    def stop_and_clear_queue(self) -> None:
        self._enqueued = []
        self._stop_if_playing()

    def stop_and_clear_queue_if_caller(self, caller: Any) -> None:
        if caller == self.current_caller:
            self.stop_and_clear_queue()

    def clear_queue_and_maybe_interrupt(self) -> None:
        self._enqueued = []
        if self.interrupt_current_audio:
            self._stop_if_playing()

    def play_file(self, filename: str) -> None:
        """Play the provided path.

        SECURITY: Filename may be an arbitrary path. For filenames coming from a collection,
        you should only ever use the os.path.basename(filename) as the filename."""
        self.play_tags([SoundOrVideoTag(filename=filename)])

    def play_file_with_caller(self, filename: str, caller: Any) -> None:
        """Play the provided path, noting down the caller.

        SECURITY: Filename may be an arbitrary path. For filenames coming from a collection,
        you should only ever use the os.path.basename(filename) as the filename."""
        if self.current_caller:
            self.current_caller_interrupted = True
        self.current_caller = caller
        self.play_file(filename)

    def insert_file(self, filename: str) -> None:
        """Place the provided path at the top of the playlist.

        SECURITY: Filename may be an arbitrary path. For filenames coming from a collection,
        you should only ever use the os.path.basename(filename) as the filename."""
        self._enqueued.insert(0, SoundOrVideoTag(filename=filename))
        self._play_next_if_idle()

    def toggle_pause(self) -> None:
        if self.current_player:
            self.current_player.toggle_pause()

    def seek_relative(self, secs: int) -> None:
        if self.current_player:
            self.current_player.seek_relative(secs)

    def shutdown(self) -> None:
        self.stop_and_clear_queue()
        for player in self.players:
            player.shutdown()
        self.players.clear()

    def _stop_if_playing(self) -> None:
        if self.current_player:
            self.current_player.stop()

    def _pop_next(self) -> AVTag | None:
        if not self._enqueued:
            return None
        return self._enqueued.pop(0)

    def _on_play_finished(self) -> None:
        if not self.current_caller_interrupted:
            self.current_caller = None
        self.current_caller_interrupted = False
        gui_hooks.av_player_did_end_playing(self.current_player)
        self.current_player = None
        self._play_next_if_idle()

    def _play_next_if_idle(self) -> None:
        if self.current_player:
            return

        next = self._pop_next()
        if next is not None:
            self._play(next)

    def _play(self, tag: AVTag) -> None:
        best_player = self._best_player_for_tag(tag)
        if best_player:
            self.current_player = best_player
            gui_hooks.av_player_will_play(tag)
            self.current_player.play(tag, self._on_play_finished)
        else:
            tooltip(f"no players found for {tag}")

    def _best_player_for_tag(self, tag: AVTag) -> Player | None:
        ranked = []
        for p in self.players:
            rank = p.rank_for_tag(tag)
            if rank is not None:
                ranked.append((rank, p))

        ranked.sort(key=itemgetter(0))

        if ranked:
            return ranked[-1][1]
        else:
            return None


av_player = AVPlayer()

# Packaged commands
##########################################################################


# return modified command array that points to bundled command, and return
# required environment
def _packagedCmd(cmd: list[str]) -> tuple[Any, dict[str, str]]:
    cmd = cmd[:]
    env = os.environ.copy()
    # keep LD_LIBRARY_PATH when in snap environment
    if "LD_LIBRARY_PATH" in env and "SNAP" not in env:
        del env["LD_LIBRARY_PATH"]

    # Try to find binary in anki-audio package for Windows/Mac
    if is_win or is_mac:
        try:
            import anki_audio

            audio_pkg_path = Path(anki_audio.__file__).parent
            if is_win:
                packaged_path = audio_pkg_path / (cmd[0] + ".exe")
            else:  # is_mac
                packaged_path = audio_pkg_path / cmd[0]

            if packaged_path.exists():
                cmd[0] = str(packaged_path)
                return cmd, env
        except ImportError:
            # anki-audio not available, fall back to old behavior
            pass

    packaged_path = Path(sys.prefix) / cmd[0]
    if packaged_path.exists():
        cmd[0] = str(packaged_path)

    return cmd, env


# Platform hacks
##########################################################################

# legacy global for add-ons
si = startup_info()


# osx throws interrupted system call errors frequently
def retryWait(proc: subprocess.Popen) -> int:
    while 1:
        try:
            return proc.wait()
        except OSError:
            continue


# Simple player implementations
##########################################################################


class SimpleProcessPlayer(Player):
    "A player that invokes a new process for each tag to play."

    args: list[str] = []
    env: dict[str, str] | None = None

    def __init__(self, taskman: TaskManager, media_folder: str | None = None) -> None:
        self._taskman = taskman
        self._media_folder = media_folder
        self._terminate_flag = False
        self._process: subprocess.Popen | None = None
        self._warned_about_missing_player = False

    def play(self, tag: AVTag, on_done: OnDoneCallback) -> None:
        self._terminate_flag = False
        self._taskman.run_in_background(
            lambda: self._play(tag),
            lambda res: self._on_done(res, on_done),
            uses_collection=False,
        )

    def stop(self) -> None:
        self._terminate_flag = True

    # note: mplayer implementation overrides this
    def _play(self, tag: AVTag) -> None:
        assert isinstance(tag, SoundOrVideoTag)
        self._process = subprocess.Popen(
            self.args + ["--", tag.path(self._media_folder)],
            env=self.env,
            cwd=self._media_folder,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        self._wait_for_termination(tag)

    def _wait_for_termination(self, tag: AVTag) -> None:
        self._taskman.run_on_main(
            lambda: gui_hooks.av_player_did_begin_playing(self, tag)
        )

        while True:
            # should we abort playing?
            if self._terminate_flag:
                self._process.terminate()
                self._process.wait(1)
                try:
                    if self._process.stdin:
                        self._process.stdin.close()
                except Exception as e:
                    print("unable to close stdin:", e)
                self._process = None
                return

            # wait for completion
            try:
                self._process.wait(0.1)
                if self._process.returncode != 0:
                    print(f"player got return code: {self._process.returncode}")
                try:
                    if self._process.stdin:
                        self._process.stdin.close()
                except Exception as e:
                    print("unable to close stdin:", e)
                self._process = None
                return
            except subprocess.TimeoutExpired:
                # process still running, repeat loop
                pass

    def _on_done(self, ret: Future, cb: OnDoneCallback) -> None:
        try:
            ret.result()
        except FileNotFoundError:
            if not self._warned_about_missing_player:
                showWarning(tr.media_sound_and_video_on_cards_will())
                self._warned_about_missing_player = True
            # must call cb() here, as we don't currently have another way
            # to flag to av_player that we've stopped
        cb()


class SimpleMpvPlayer(SimpleProcessPlayer, VideoPlayer):
    default_rank = 1

    args, env = _packagedCmd(
        [
            "mpv",
            "--no-terminal",
            "--force-window=no",
            "--ontop",
            "--audio-display=no",
            "--keep-open=no",
            "--input-media-keys=no",
            "--autoload-files=no",
            "--no-ytdl",
        ]
    )

    def __init__(
        self, taskman: TaskManager, base_folder: str, media_folder: str
    ) -> None:
        super().__init__(taskman, media_folder)
        self.args += [f"--config-dir={base_folder}"]


class SimpleMplayerPlayer(SimpleProcessPlayer, SoundOrVideoPlayer):
    args, env = _packagedCmd(["mplayer", "-really-quiet", "-noautosub"])
    if is_win:
        args += ["-ao", "win32"]


# MPV
##########################################################################


class MpvManager(MPV, SoundOrVideoPlayer):
    if not is_lin:
        default_argv = MPVBase.default_argv + [
            "--input-media-keys=no",
        ]

    def __init__(self, base_path: str, media_folder: str) -> None:
        self.media_folder = media_folder
        mpvPath, self.popenEnv = _packagedCmd(["mpv"])
        self.executable = mpvPath[0]
        self._on_done: OnDoneCallback | None = None
        self.default_argv += [f"--config-dir={base_path}"]
        super().__init__(window_id=None, debug=False)

    def on_init(self) -> None:
        # if mpv dies and is restarted, tell Anki the
        # current file is done
        if self._on_done:
            self._on_done()

        m = re.search(r"(\d+)\.(\d+)\.(\d+)", self.get_property("mpv-version"))
        if m:
            self.mpv_version = (int(m[1]), int(m[2]), int(m[3]))
        else:
            self.mpv_version = None

        try:
            self.command("keybind", "q", "stop")
            self.command("keybind", "Q", "stop")
            self.command("keybind", "CLOSE_WIN", "stop")
            self.command("keybind", "ctrl+w", "stop")
            self.command("keybind", "ctrl+c", "stop")
        except MPVCommandError:
            print("mpv too old for key rebinding")

    def play(self, tag: AVTag, on_done: OnDoneCallback) -> None:
        assert isinstance(tag, SoundOrVideoTag)
        self._on_done = on_done
        path = tag.path(self.media_folder)

        if self.mpv_version is None or self.mpv_version >= (0, 38, 0):
            self.command("loadfile", path, "replace", -1, "pause=no")
        else:
            self.command("loadfile", path, "replace", "pause=no")
        gui_hooks.av_player_did_begin_playing(self, tag)

    def stop(self) -> None:
        self.command("stop")

    def toggle_pause(self) -> None:
        self.command("cycle", "pause")

    def seek_relative(self, secs: int) -> None:
        self.command("seek", secs, "relative")

    def on_property_idle_active(self, value: bool) -> None:
        if value and self._on_done:
            from aqt import mw

            mw.taskman.run_on_main(self._on_done)

    def shutdown(self) -> None:
        self.close()

    # Legacy, not used
    ##################################################

    togglePause = toggle_pause
    seekRelative = seek_relative

    def queueFile(self, file: str) -> None:
        return

    def clearQueue(self) -> None:
        return


# Mplayer in slave mode
##########################################################################


class SimpleMplayerSlaveModePlayer(SimpleMplayerPlayer):
    def __init__(self, taskman: TaskManager, media_folder: str) -> None:
        self.media_folder = media_folder
        super().__init__(taskman, media_folder)
        self.args.append("-slave")

    def _play(self, tag: AVTag) -> None:
        assert isinstance(tag, SoundOrVideoTag)

        self._process = subprocess.Popen(
            self.args + ["--", tag.path(self.media_folder)],
            env=self.env,
            cwd=self.media_folder,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            startupinfo=startup_info(),
        )
        self._wait_for_termination(tag)

    def command(self, *args: Any) -> None:
        """Send a command over the slave interface.

        The trailing newline is automatically added."""
        str_args = [str(x) for x in args]
        if self._process:
            self._process.stdin.write(" ".join(str_args).encode("utf8") + b"\n")
            self._process.stdin.flush()

    def seek_relative(self, secs: int) -> None:
        self.command("seek", secs, 0)

    def toggle_pause(self) -> None:
        self.command("pause")


# MP3 transcoding
##########################################################################


def _encode_mp3(src_wav: str, dst_mp3: str) -> None:
    cmd = ["lame", src_wav, dst_mp3, "--noreplaygain", "--quiet"]
    cmd, env = _packagedCmd(cmd)
    try:
        retcode = retryWait(subprocess.Popen(cmd, startupinfo=startup_info(), env=env))
    except Exception as e:
        raise Exception(tr.media_error_running(val=" ".join(cmd))) from e
    if retcode != 0:
        raise Exception(tr.media_error_running(val=" ".join(cmd)))

    os.unlink(src_wav)


def encode_mp3(mw: aqt.AnkiQt, src_wav: str, on_done: Callable[[str], None]) -> None:
    "Encode the provided wav file to .mp3, and call on_done() with the path."
    dst_mp3 = src_wav.replace(".wav", "%d.mp3" % time.time())

    def _on_done(fut: Future) -> None:
        if exc := fut.exception():
            print(exc)
            showWarning(tr.editing_couldnt_record_audio_have_you_installed())
            return

        on_done(dst_mp3)

    mw.taskman.run_in_background(
        lambda: _encode_mp3(src_wav, dst_mp3), _on_done, uses_collection=False
    )


# Recording interface
##########################################################################


class Recorder(ABC):
    # seconds to wait before recording
    STARTUP_DELAY = 0.3

    def __init__(self, output_path: str) -> None:
        self.output_path = output_path

    def start(self, on_done: Callable[[], None]) -> None:
        "Start recording, then call on_done() when started."
        self._started_at = time.time()
        on_done()

    def stop(self, on_done: Callable[[str], None]) -> None:
        "Stop recording, then call on_done() when finished."
        on_done(self.output_path)

    def duration(self) -> float:
        "Seconds since recording started."
        return time.time() - self._started_at

    def on_timer(self) -> None:
        "Will be called periodically."


# QAudioInput recording
##########################################################################


class QtAudioInputRecorder(Recorder):
    def __init__(self, output_path: str, mw: aqt.AnkiQt, parent: QWidget) -> None:
        super().__init__(output_path)

        self.mw = mw
        self._parent = parent

        from PyQt6.QtMultimedia import QAudioSource, QMediaDevices  # type: ignore

        # Get the default audio input device
        device = QMediaDevices.defaultAudioInput()

        # Try to use Int16 format first (avoids conversion)
        preferred_format = device.preferredFormat()
        int16_format = preferred_format
        int16_format.setSampleFormat(preferred_format.SampleFormat.Int16)

        if device.isFormatSupported(int16_format):
            # Use Int16 if supported
            format = int16_format
        else:
            # Fall back to device's preferred format
            format = preferred_format

        # Create the audio source with the chosen format
        source = QAudioSource(device, format, parent)

        # Store the actual format being used
        self._format = source.format()
        self._audio_input = source

    def _convert_float_to_int16(self, float_buffer: bytearray) -> bytes:
        """Convert float32 audio samples to int16 format for WAV output."""
        import struct

        float_count = len(float_buffer) // 4  # 4 bytes per float32
        floats = struct.unpack(f"{float_count}f", float_buffer)

        # Convert to int16 range, clipping and scaling in one step
        int16_samples = [
            max(-32768, min(32767, int(max(-1.0, min(1.0, f)) * 32767))) for f in floats
        ]

        return struct.pack(f"{len(int16_samples)}h", *int16_samples)

    def start(self, on_done: Callable[[], None]) -> None:
        self._iodevice = self._audio_input.start()
        self._buffer = bytearray()
        qconnect(self._iodevice.readyRead, self._on_read_ready)
        super().start(on_done)

    def _on_read_ready(self) -> None:
        self._buffer.extend(cast(bytes, self._iodevice.readAll()))

    def stop(self, on_done: Callable[[str], None]) -> None:
        from PyQt6.QtMultimedia import QAudio

        def on_stop_timer() -> None:
            # read anything remaining in buffer & stop
            self._on_read_ready()
            self._audio_input.stop()

            if (err := self._audio_input.error()) != QAudio.Error.NoError:
                showWarning(f"recording failed: {err}")
                return

            def write_file() -> None:
                from PyQt6.QtMultimedia import QAudioFormat

                # swallow the first 300ms to allow audio device to quiesce
                bytes_per_frame = self._format.bytesPerFrame()
                frames_to_skip = int(self._format.sampleRate() * self.STARTUP_DELAY)
                bytes_to_skip = frames_to_skip * bytes_per_frame

                if len(self._buffer) <= bytes_to_skip:
                    return
                self._buffer = self._buffer[bytes_to_skip:]

                # Check if we need to convert float samples to int16
                if self._format.sampleFormat() == QAudioFormat.SampleFormat.Float:
                    audio_data = self._convert_float_to_int16(self._buffer)
                    sample_width = 2  # int16 is 2 bytes
                else:
                    # For integer formats, use the data as-is
                    audio_data = bytes(self._buffer)
                    sample_width = self._format.bytesPerSample()

                # write out the wave file with the correct format parameters
                wf = wave.open(self.output_path, "wb")
                wf.setnchannels(self._format.channelCount())
                wf.setsampwidth(sample_width)
                wf.setframerate(self._format.sampleRate())
                wf.writeframes(audio_data)
                wf.close()

            def and_then(fut: Future) -> None:
                fut.result()
                Recorder.stop(self, on_done)

            self.mw.taskman.run_in_background(
                write_file, and_then, uses_collection=False
            )

        # schedule the stop for half a second in the future,
        # to avoid truncating the end of the recording
        self._stop_timer = t = QTimer(self._parent)
        t.timeout.connect(on_stop_timer)  # type: ignore
        t.setSingleShot(True)
        t.start(500)


# Native macOS recording
##########################################################################


class NativeMacRecorder(Recorder):
    def __init__(self, output_path: str) -> None:
        super().__init__(output_path)
        self._error: str | None = None

    def _on_error(self, msg: str) -> None:
        self._error = msg

    def start(self, on_done: Callable[[], None]) -> None:
        self._error = None
        assert macos_helper
        macos_helper.start_wav_record(self.output_path, self._on_error)
        super().start(on_done)

    def stop(self, on_done: Callable[[str], None]) -> None:
        assert macos_helper
        macos_helper.end_wav_record()
        Recorder.stop(self, on_done)


# Recording dialog
##########################################################################


class RecordDialog(QDialog):
    _recorder: Recorder

    def __init__(
        self,
        parent: QWidget,
        mw: aqt.AnkiQt,
        on_success: Callable[[str], None],
    ):
        QDialog.__init__(self, parent)
        self._parent = parent
        self.mw = mw
        self._on_success = on_success
        disable_help_button(self)

        self._start_recording()
        self._setup_dialog()

    def _setup_dialog(self) -> None:
        self.setWindowTitle("Anki")
        icon = QLabel()
        qicon = theme_manager.icon_from_resources("icons:media-record.svg")
        icon.setPixmap(qicon.pixmap(60, 60))
        self.label = QLabel("...")
        hbox = QHBoxLayout()
        hbox.addWidget(icon)
        hbox.addWidget(self.label)
        v = QVBoxLayout()
        v.addLayout(hbox)
        buts = (
            QDialogButtonBox.StandardButton.Save
            | QDialogButtonBox.StandardButton.Cancel
        )
        b = QDialogButtonBox(buts)  # type: ignore
        v.addWidget(b)
        self.setLayout(v)
        save_button = b.button(QDialogButtonBox.StandardButton.Save)
        save_button.setDefault(True)
        save_button.setAutoDefault(True)
        qconnect(save_button.clicked, self.accept)
        cancel_button = b.button(QDialogButtonBox.StandardButton.Cancel)
        cancel_button.setDefault(False)
        cancel_button.setAutoDefault(False)
        qconnect(cancel_button.clicked, self.reject)
        restoreGeom(self, "audioRecorder2")
        self.show()

    def _save_diag(self) -> None:
        saveGeom(self, "audioRecorder2")

    def _start_recording(self) -> None:
        if macos_helper and platform.machine() == "arm64":
            self._recorder = NativeMacRecorder(
                namedtmp("rec.wav"),
            )
        else:
            self._recorder = QtAudioInputRecorder(
                namedtmp("rec.wav"), self.mw, self._parent
            )
        self._recorder.start(self._start_timer)

    def _start_timer(self) -> None:
        self._timer = t = QTimer(self._parent)
        t.timeout.connect(self._on_timer)  # type: ignore
        t.setSingleShot(False)
        t.start(100)

    def _on_timer(self) -> None:
        self._recorder.on_timer()
        duration = self._recorder.duration()
        self.label.setText(tr.media_recordingtime(secs=f"{duration:0.1f}"))

    def accept(self) -> None:
        self._timer.stop()

        try:
            self._save_diag()
            self._recorder.stop(self._on_success)
        finally:
            QDialog.accept(self)

    def reject(self) -> None:
        self._timer.stop()

        def cleanup(out: str) -> None:
            os.unlink(out)

        try:
            self._recorder.stop(cleanup)
        finally:
            QDialog.reject(self)


def record_audio(
    parent: QWidget, mw: aqt.AnkiQt, encode: bool, on_done: Callable[[str], None]
) -> None:
    def after_record(path: str) -> None:
        if not encode:
            on_done(path)
        else:
            encode_mp3(mw, path, on_done)

    try:
        _diag = RecordDialog(parent, mw, after_record)
    except Exception as e:
        err_str = str(e)
        showWarning(markdown(tr.qt_misc_unable_to_record(error=err_str)))


# Legacy audio interface
##########################################################################
# these will be removed in the future


def clearAudioQueue() -> None:
    av_player.stop_and_clear_queue()


def play(filename: str) -> None:
    av_player.play_file(filename)


def playFromText(text: Any) -> None:
    print("playFromText() deprecated")


# legacy globals
_player = play
_queueEraser = clearAudioQueue
mpvManager: MpvManager | None = None

# add everything from this module into anki.sound for backwards compat
_exports = [i for i in locals().items() if not i[0].startswith("__")]
for k, v in _exports:
    sys.modules["anki.sound"].__dict__[k] = v

# Tag handling
##########################################################################


def av_refs_to_play_icons(text: str) -> str:
    """Add play icons into the HTML.

    When clicked, the icon will call eg pycmd('play:q:1').
    """

    def repl(match: re.Match) -> str:
        return f"""
<a class="replay-button soundLink" href=# onclick="pycmd('{match.group(1)}'); return false;" draggable="false">
    <svg class="playImage" viewBox="0 0 64 64" version="1.1">
        <circle cx="32" cy="32" r="29" />
        <path d="M56.502,32.301l-37.502,20.101l0.329,-40.804l37.173,20.703Z" />
    </svg>
</a>"""

    return AV_REF_RE.sub(repl, text)


def play_clicked_audio(pycmd: str, card: Card) -> None:
    """eg. if pycmd is 'play:q:0', play the first audio on the question side."""
    play, context, str_idx = pycmd.split(":")
    idx = int(str_idx)
    if context == "q":
        tags = card.question_av_tags()
    else:
        tags = card.answer_av_tags()
    av_player.play_tags([tags[idx]])


# Init defaults
##########################################################################


def setup_audio(taskman: TaskManager, base_folder: str, media_folder: str) -> None:
    # legacy global var
    global mpvManager

    try:
        mpvManager = MpvManager(base_folder, media_folder)
    except FileNotFoundError:
        print("mpv not found, reverting to mplayer")
    except aqt.mpv.MPVProcessError:
        print(traceback.format_exc())
        print("mpv too old or failed to open, reverting to mplayer")

    if mpvManager is not None:
        av_player.players.append(mpvManager)

        if is_win:
            mpvPlayer = SimpleMpvPlayer(taskman, base_folder, media_folder)
            av_player.players.append(mpvPlayer)
    else:
        mplayer = SimpleMplayerSlaveModePlayer(taskman, media_folder)
        av_player.players.append(mplayer)

    # tts support
    if is_mac:
        from aqt.tts import MacTTSPlayer

        av_player.players.append(MacTTSPlayer(taskman))
    elif is_win:
        from aqt.tts import WindowsTTSPlayer

        av_player.players.append(WindowsTTSPlayer(taskman))

        if platform.release() == "10":
            from aqt.tts import WindowsRTTTSFilePlayer

            # If Windows 10, ensure it's October 2018 update or later
            if int(platform.version().split(".")[-1]) >= 17763:
                av_player.players.append(WindowsRTTTSFilePlayer(taskman))


def cleanup_audio() -> None:
    av_player.shutdown()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Callable
from enum import Enum, auto

import aqt
import aqt.browser
import aqt.gui_hooks
from aqt.qt import *
from aqt.theme import theme_manager
from aqt.utils import tr


class SidebarTool(Enum):
    SELECT = auto()
    SEARCH = auto()


class SidebarToolbar(QToolBar):
    _tools: tuple[tuple[SidebarTool, str, Callable[[], str]], ...] = (
        (
            SidebarTool.SEARCH,
            "mdi:magnify",
            tr.actions_search,
        ),
        (
            SidebarTool.SELECT,
            "mdi:selection-drag",
            tr.actions_select,
        ),
    )

    def __init__(self, sidebar: aqt.browser.sidebar.SidebarTreeView) -> None:
        super().__init__()
        self.sidebar = sidebar
        self._action_group = QActionGroup(self)
        qconnect(self._action_group.triggered, self._on_action_group_triggered)
        self._setup_tools()
        self.setIconSize(QSize(18, 18))
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setStyle(QStyleFactory.create("fusion"))
        aqt.gui_hooks.theme_did_change.append(self._update_icons)

    def _setup_tools(self) -> None:
        for row, tool in enumerate(self._tools):
            action = self.addAction(
                theme_manager.icon_from_resources(tool[1]), tool[2]()
            )
            assert action is not None
            action.setCheckable(True)
            action.setShortcut(f"Alt+{row + 1}")
            self._action_group.addAction(action)
        # always start with first tool
        active = 0
        self._action_group.actions()[active].setChecked(True)
        self.sidebar.tool = self._tools[active][0]

    def _on_action_group_triggered(self, action: QAction) -> None:
        index = self._action_group.actions().index(action)
        self.sidebar.tool = self._tools[index][0]

    def cleanup(self) -> None:
        aqt.gui_hooks.theme_did_change.remove(self._update_icons)

    def _update_icons(self) -> None:
        for idx, action in enumerate(self._action_group.actions()):
            action.setIcon(theme_manager.icon_from_resources(self._tools[idx][1]))



# Form implementation generated from reading ui file 'qt/aqt/forms/changemodel.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(362, 391)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setSpacing(10)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setVerticalSpacing(4)
        self.gridLayout.setObjectName("gridLayout")
        self.label_6 = QtWidgets.QLabel(parent=Dialog)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.oldModelLabel = QtWidgets.QLabel(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.oldModelLabel.sizePolicy().hasHeightForWidth())
        self.oldModelLabel.setSizePolicy(sizePolicy)
        self.oldModelLabel.setText("")
        self.oldModelLabel.setObjectName("oldModelLabel")
        self.gridLayout.addWidget(self.oldModelLabel, 0, 1, 1, 1)
        self.label = QtWidgets.QLabel(parent=Dialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)
        self.modelChooserWidget = QtWidgets.QWidget(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.modelChooserWidget.sizePolicy().hasHeightForWidth())
        self.modelChooserWidget.setSizePolicy(sizePolicy)
        self.modelChooserWidget.setObjectName("modelChooserWidget")
        self.gridLayout.addWidget(self.modelChooserWidget, 1, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.tgroup = QtWidgets.QGroupBox(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tgroup.sizePolicy().hasHeightForWidth())
        self.tgroup.setSizePolicy(sizePolicy)
        self.tgroup.setObjectName("tgroup")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tgroup)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.scrollArea = QtWidgets.QScrollArea(parent=self.tgroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.templateMap = QtWidgets.QWidget()
        self.templateMap.setGeometry(QtCore.QRect(0, 0, 330, 120))
        self.templateMap.setObjectName("templateMap")
        self.scrollArea.setWidget(self.templateMap)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.verticalLayout.addWidget(self.tgroup)
        self.fgroup = QtWidgets.QGroupBox(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fgroup.sizePolicy().hasHeightForWidth())
        self.fgroup.setSizePolicy(sizePolicy)
        self.fgroup.setObjectName("fgroup")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.fgroup)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.scrollArea_2 = QtWidgets.QScrollArea(parent=self.fgroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea_2.sizePolicy().hasHeightForWidth())
        self.scrollArea_2.setSizePolicy(sizePolicy)
        self.scrollArea_2.setWidgetResizable(True)
        self.scrollArea_2.setObjectName("scrollArea_2")
        self.fieldMap = QtWidgets.QWidget()
        self.fieldMap.setGeometry(QtCore.QRect(0, 0, 330, 119))
        self.fieldMap.setObjectName("fieldMap")
        self.scrollArea_2.setWidget(self.fieldMap)
        self.verticalLayout_3.addWidget(self.scrollArea_2)
        self.verticalLayout.addWidget(self.fgroup)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Help|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.browsing_change_note_type())
        self.label_6.setText(tr.browsing_current_note_type())
        self.label.setText(tr.browsing_new_note_type())
        self.tgroup.setTitle(tr.editing_cards())
        self.fgroup.setTitle(tr.editing_fields())